# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
# Usage subject to license agreement
# hello@dativa.com for more information

from dativa.scrubber.tests import _BaseTest
from dativa.scrubber import ScrubberValidationError


class StringTests(_BaseTest):

    def test_string(self):
        self._test_filter(dirty_file="string/test_string_is_unique_dirty.csv",
                          clean_file="string/test_string_is_unique_clean.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "default_value": "N/A",
                                          "fallback_mode": "remove_record",

                                          "is_unique": True,
                                          "lookalike_match": False,
                                          "maximum_length": 15,
                                          "minimum_length": 0,
                                          "regex": "",
                                          "skip_blank": False
                                      },
                                      "rule_type": "String",
                                      "field": "Name"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Name',
                                  'rule': 'String',
                                  'number_records': 5,
                                  'category': 'quarantined',
                                  'description': 'Removed duplicates',
                                  'df': [['Sarvesh'],
                                         ['Sarvesh'],
                                         ['Narendra'],
                                         ['Harshad'],
                                         ['sivaji']],

                              },
                          ]
                          )

    def test_string_2(self):
        self._test_filter(dirty_file="string/test_string_dirty.csv",
                          clean_file="string/test_string_clean.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "default_value": "N/A",
                                          "fallback_mode": "use_default",

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_length": 15,
                                          "minimum_length": 0,
                                          "regex": "^[A-z\\s]{1,}$",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "String",
                                      "field": "Name"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Name',
                                  'rule': 'String',
                                  'number_records': 10,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [['Loida Lajeunesse', 'N/A'],
                                         ['Michael Markwell', 'N/A'],
                                         ['Iris Illingworth', 'N/A'],
                                         ['Samatha Severance', 'N/A'],
                                         ['Anneliese Angers', 'N/A'],
                                         ['Polly Porterfield', 'N/A'],
                                         ['Bernadette Bratt', 'N/A'],
                                         ['Margarito Molloy', 'N/A'],
                                         ['FerNonede Fuentes', 'N/A'],
                                         ['Kurtis K3en', 'N/A']],

                              },
                          ])

    def test_string_3(self):
        self._test_filter(dirty_file="string/test_string_dirty1.csv",
                          clean_file="string/test_string_clean1.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "default_value": "N/A",
                                          "fallback_mode": "use_default",

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_length": 15,
                                          "minimum_length": 0,
                                          "regex": ".+",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "String",
                                      "field": "Name"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Name',
                                  'rule': 'String',
                                  'number_records': 12,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [['Loida Lajeunesse', 'N/A'],
                                         [None, 'N/A'],
                                         ['Michael Markwell', 'N/A'],
                                         ['Iris Illingworth', 'N/A'],
                                         ['Samatha Severance', 'N/A'],
                                         ['Anneliese Angers', 'N/A'],
                                         ['Polly Porterfield', 'N/A'],
                                         ['Bernadette Bratt', 'N/A'],
                                         [None, 'N/A'],
                                         ['Margarito Molloy', 'N/A'],
                                         ['FerNonede Fuentes', 'N/A'],
                                         [None, 'N/A']],

                              },
                          ])

    def test_string_4(self):
        self._test_filter(dirty_file="string/test_string_dirty.csv",
                          clean_file="string/test_string_clean2.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "default_value": "N/A",
                                          "fallback_mode": "remove_record",

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_length": 15,
                                          "minimum_length": 0,
                                          "regex": "",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "String",
                                      "field": "Name"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Name',
                                  'rule': 'String',
                                  'number_records': 9,
                                  'category': 'quarantined',
                                  'description': 'Data quarantined',
                                  'df': [['Loida Lajeunesse'],
                                         ['Michael Markwell'],
                                         ['Iris Illingworth'],
                                         ['Samatha Severance'],
                                         ['Anneliese Angers'],
                                         ['Polly Porterfield'],
                                         ['Bernadette Bratt'],
                                         ['Margarito Molloy'],
                                         ['FerNonede Fuentes']],

                              },
                          ])

    def test_string_ints(self):
        self._test_filter(dirty_file="string/test_string_dirty_ints.csv",
                          clean_file="string/test_string_clean_ints.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "fallback_mode": "remove_record",
                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_length": 10,  # but we're filtering with regex
                                          "minimum_length": 0,
                                          "regex": "^\d$",  # number, length 1
                                          "skip_blank": False,
                                      },
                                      "rule_type": "String",
                                      "field": "integer"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'integer',
                                  'rule': 'String',
                                  'number_records': 1,
                                  'category': 'quarantined',
                                  'description': 'Data quarantined',
                                  'df': [[10]],

                              },
                          ])

    def test_string_FpValidationError(self):
        self._test_filter(dirty_file="string/test_string_dirty.csv",
                          clean_file="string/test_string_clean2.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "default_value": "N/A",
                                          "fallback_mode": "remove_record",

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_length": 15,
                                          "minimum_length": 0,
                                          "regex": "**",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "String",
                                      "field": "Name"
                                  }
                              ]
                          },
                          expected_error=ScrubberValidationError)

    def test_string_FpValidationError2(self):
        self._test_filter(dirty_file="string/test_string_dirty.csv",
                          clean_file="string/test_string_clean2.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "default_value": "N/A",
                                          "fallback_mode": "remove_record",
                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_length": 15,
                                          "minimum_length": 0,
                                          "regex": "*",
                                          "not_a_field": True,
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "String",
                                      "field": "Name"
                                  }
                              ]
                          },
                          expected_error=ScrubberValidationError)

    def test_string_5(self):
        self._test_filter(dirty_file="string/test_string_dirty.csv",
                          clean_file="string/test_string_clean_removed_all.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "default_value": "N/A",
                                          "fallback_mode": "remove_record",

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_length": 15,
                                          "minimum_length": 0,
                                          "regex": ".*",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "String",
                                      "field": "NameNotThere"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'NameNotThere',
                                  'rule': 'String',
                                  'number_records': 50,
                                  'category': 'quarantined',
                                  'description': 'File did not contain required column',
                                  'df': [['Glinda Geisler', None],
                                         ['Loida Lajeunesse', None],
                                         ['Dani Deboer', None],
                                         ['Trena Teti', None],
                                         ['Vanda Viers', None],
                                         ['Eveline Eddy', None],
                                         ['Zora Zajac', None],
                                         ['Nicole Nickens', None],
                                         ['Suzy Stetler', None],
                                         ['Karine Kissee', None],
                                         ['Deidre Durrell', None],
                                         ['Nery Novy', None],
                                         ['Michael Markwell', None],
                                         ['FerNoneda Faison', None],
                                         ['Tyler Thrall', None],
                                         ['Tianna Tenner', None],
                                         ['Iris Illingworth', None],
                                         ['Samatha Severance', None],
                                         ['Anneliese Angers', None],
                                         ['Polly Porterfield', None],
                                         ['Bernadette Bratt', None],
                                         ['Siu Siddiqi', None],
                                         ['Susan Salmon', None],
                                         ['Dexter Desch', None],
                                         ['Amber Avis', None],
                                         ['Jackson Jetton', None],
                                         ['Vernia Vue', None],
                                         ['Sanda Sandoval', None],
                                         ['Margarito Molloy', None],
                                         ['Timothy Tranmer', None],
                                         ['Gertie Godin', None],
                                         ['Jenae Jack', None],
                                         ['FerNonede Fuentes', None],
                                         ['Yaeko Yoshimura', None],
                                         ['Lovetta Lowrey', None],
                                         ['Eartha Ellisor', None],
                                         ['Elane Edds', None],
                                         ['Emilie Ecker', None],
                                         ['Annette Alcaraz', None],
                                         ['Larue Liptak', None],
                                         ['Talia Tugwell', None],
                                         ['Sachiko Suniga', None],
                                         ['Les Lieser', None],
                                         ['Emeline Ekberg', None],
                                         ['Whitney Wymore', None],
                                         ['Kurtis K3en', None],
                                         ['Lloyd Landa', None],
                                         ['Gregorio Gorder', None],
                                         ['Chas Carneva', None],
                                         ['Lashaunda Lindl', None]],

                              },
                          ])

    def test_string_FpValidationError_2(self):
        self._test_filter(dirty_file="string/test_string_dirty.csv",
                          clean_file="string/test_string_clean2.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "default_value": "N/A",
                                          "fallback_mode": "remove_record",
                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_length": 15,
                                          "minimum_length": 20,
                                          "regex": ".*",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "String",
                                      "field": "Name"
                                  }
                              ]
                          },
                          expected_error=ScrubberValidationError)



# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
#  This code is licensed under MIT license (see license.txt for details)

import pandas as pd
from dativa.tools.pandas import get_column_name
from .report import DefaultReportWriter


class FallbackType:
    """
    Specifies how to handle data that is not matched
    """

    USE_DEFAULT_VALUE = "use_default"
    USE_INVALID_DATA = "do_not_replace"
    REMOVE_ENTRY = "remove_record"

    def pretty(self):
        if self == FallbackType.USE_DEFAULT_VALUE:
            return "Replaced with default value"
        elif self == FallbackType.USE_INVALID_DATA:
            return "No changes made"
        elif self == FallbackType.REMOVE_ENTRY:
            return "Data quarantined"


class DateRangeType:
    """
    Specifies how data ranges should be checked
    """
    NONE = "none"
    FIXED = "fixed"
    ROLLING = "rolling"


class UniqueFieldsType:
    """
    A specific type of rule for comparing data using a duplication rule
    """

    CHECK_THIS_COLUMN_ONLY = "check_this_column_only"
    CHECK_OTHER_COLUMNS_ONLY = "check_other_columns"
    CHECK_ALL_COLUMNS = "check_all_columns"


class SessionOverlapsType:
    """
    Specifies how session overlaps should be handled
    """

    IGNORE = "ignore"
    TRUNCATE_START = "truncate_start"
    TRUNCATE_END = "truncate_end"


class SessionGapsType:
    """
    Specifies how session gaps should be handled
    """

    IGNORE = "ignore"
    EXTEND_START = "extend_start"
    EXTEND_END = "extend_end"
    INSERT_NEW = "insert_new"


class HistoryType:
    """
    Specified different types of category of error
    that can be logged to a history object
    """

    IGNORED = 'ignored'
    REPLACED = 'replaced'
    REMOVED = 'quarantined'
    MODIFIED = 'modified'
    INSERTED = 'inserted'


"""
A set of error classes raised by the FileProcessing Modules
"""


class ScrubberError(Exception):
    def __init__(self, message="File Processing Error"):
        self.message = message


class ScrubberTooManyRecordsError(Exception):
    def __init__(self, message="Too Many Records"):
        self.message = message


class ScrubberValidationError(Exception):
    def __init__(self, message="Invalid configuration"):
        self.message = message


class DefaultProfile():
    maximum_records = 2000000
    maximum_records_closest_matches = 5000
    lookalike_number_records = 5
    maximum_records_lookalike = 5000
    maximum_file_records_lookalike = 500000


class BaseAddIn:

    def __init__(self, validator):
        self.validator = validator
        self.report_writer = validator.report_writer


class BaseValidator:
    """
    The base model for Validators that are run.

    Validators must have the following:
    rule_type: a string name
    fields: a dictionary of fields, including default values and validation functions
    run(df, column=None): a function to run the validator in-situe on the passed DataFrame

    Parameters
    ----------
    params: a dict of parameters for the validator
    report_writer: a report writer class for logging
    df_dict: a dictionary of DataFrame referenced in the params function
    """
    rule_type = None
    fields = {}
    reference_list = None

    def _append_addin_fields(self, addins):
        for AddIn in addins:
            for field in AddIn.fields:
                self.fields[field] = AddIn.fields[field]

    def __init__(self, params=None, report_writer=None, df_dict=None, profile=None):

        # add any fields from the addins
        addins = []
        if hasattr(self, "pre_execution"):
            self._append_addin_fields(self.pre_execution)
            addins = addins + self.pre_execution

        if hasattr(self, "post_execution"):
            self._append_addin_fields(self.post_execution)
            addins = addins + self.post_execution

        if params is not None:
            # check for any parameters that are not in the fields...
            for param in params:
                if param not in self.fields:
                    raise ScrubberValidationError("The param {field} is not valid for {rule_type} rules".format(
                        field=param, rule_type=self.rule_type))

            # check for mandatory parameters
            for field in self.fields:
                if field in params:
                    self._validate(field, params[field])
                    setattr(self, field, params[field])
                elif self.fields[field]["default"] is None:
                    raise ScrubberValidationError("The param {field} must be set in {rule_type} rules".format(
                        field=field, rule_type=self.rule_type))
                else:
                    setattr(self, field, self.fields[field]["default"])

            # set up fixed variable
            if report_writer is None:
                self.report_writer = DefaultReportWriter()
            else:
                self.report_writer = report_writer
            if profile is None:
                self.profile = DefaultProfile()
            else:
                self.profile = profile
            if df_dict is None or type(df_dict) is dict:
                self.df_dict = df_dict
            else:
                raise ScrubberValidationError(
                    "Passed df_dict is not a python dict, type is {0}".format(type(df_dict)))

            # run any cross validation
            if hasattr(self, "cross_validation"):
                self.cross_validation()
            for AddIn in addins:
                add_in = AddIn(self)
                if hasattr(add_in, "cross_validation"):
                    add_in.cross_validation()

    def _validate(self, field, value):
        """
        Runs the validation function for a field in the validator
        """
        try:
            if self.fields[field]["validator"](value) is True:
                return
        except:  # noqa
            pass

        raise ScrubberValidationError("'{value}' is not valid for {param} in a {type} rule".format(
            param=field, value=value, type=self.rule_type))

    def get_df(self, name):
        """
        Returns a named DataFrame from within the df_dict dictionary passed
        when the class was instantiated
        """
        return pd.DataFrame(self.df_dict[name]).copy()

    def run_all(self, df, column=None):

        if column is not None and not self.fields.get('skip_blanks'):
            try:
                get_column_name(df, column)
            except KeyError as e:

                self.report_writer.log_history(
                    self.rule_type,
                    column,
                    df,
                    HistoryType.REMOVED,
                    "File did not contain required column")
                df[column] = None
                df.dropna(subset=[column], inplace=True)  # we already checked that this column does not exist
                df.drop(columns=[column], inplace=True)
                return  # goto: getout

        if hasattr(self, "pre_execution"):
            for AddIn in self.pre_execution:
                add_in = AddIn(self)
                add_in.run(df, get_column_name(df, column))

        if hasattr(self, "run"):
            if column is None:
                self.run(df)
            else:
                self.run(df, column)

        if hasattr(self, "post_execution"):
            for AddIn in self.post_execution:
                add_in = AddIn(self)
                if df.shape[0] > 0:  # stop processing if all records have been deleted
                    add_in.run(df, df.columns[-1], get_column_name(df, column))

    def get_dict(self, field):

        row = dict()

        row["field"] = field
        row["type"] = self.rule_type

        for field in self.fields:
            attr = getattr(self, field)
            if attr is not None and attr != self.fields[field]["default"]:
                row[field] = getattr(self, field)

        return row



# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
# Usage subject to license agreement
# hello@dativa.com for more information

import logging
import tempfile
import shutil
from datetime import datetime
import pandas as pd
from dativa.scrubber.tests import _BaseTest
from dativa.tools.pandas import CSVHandler
from dativa.scrubber import ScrubberValidationError, Scrubber, LoggingReportWriter, DefaultReportWriter, \
    PersistentFieldLogger


class WorkflowTests(_BaseTest):

    def test_report_writer(self):
        csv = CSVHandler(base_path=self.base_path)

        df = csv.get_dataframe("number/test_int_is_unique_dirty.csv", force_dtype=pd.np.float)

        writer = DefaultReportWriter()
        scrubber = Scrubber(report_writer=writer)

        report = scrubber.run(df=df,
                              config={"rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "decimal_places": 1,
                                          "default_value": 0,
                                          "fallback_mode": "use_default",
                                          "fix_decimal_places": True,

                                          "is_unique": True,
                                          "lookalike_match": False,
                                          "maximum_value": 100.0,
                                          "minimum_value": 1.0,
                                          "skip_blank": True,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Number",
                                      "field": "TotalEpisodes"
                                  }]})

        self.assertEqual(str(report[0]).split(", ")[1],
                         "Field TotalEpisodes(Number): #26 modified/|Automatically fixed to 1 decimal places")
        self.assertEqual(str(report[1]).split(", ")[1],
                         "Field TotalEpisodes_clean(Number): #16 quarantined/|Removed duplicates")
        self.assertEqual(str(report[2]).split(", ")[1],
                         "Field TotalEpisodes(Number): #1 replaced/|Replaced with default value")

        self.assertEqual(report[0].get_log_dict()["field"], "TotalEpisodes")
        self.assertEqual(report[0].get_log_dict()["rule"], "Number")
        self.assertEqual(report[0].get_log_dict()["number_records"], 26)
        self.assertEqual(report[0].get_log_dict()["category"], "modified")
        self.assertEqual(report[0].get_log_dict()["description"], "Automatically fixed to 1 decimal places")

        self.assertEqual(report[0].get_records().shape, (26, 2))

    def test_log_report_writer(self):
        csv = CSVHandler(base_path=self.base_path)
        logger = logging.getLogger("dativa.scrubber.test")
        tempdir = tempfile.mkdtemp()

        df = csv.get_dataframe("number/test_int_is_unique_dirty.csv", force_dtype=pd.np.float)

        writer = LoggingReportWriter(logger=logger,
                                     quarantine_path=tempdir,
                                     error_severities=["quarantined"],
                                     warning_severities=["replaced"])
        scrubber = Scrubber(report_writer=writer)

        with self.assertLogs("dativa.scrubber.test") as logs:
            scrubber.run(df=df,
                         config={"rules": [
                             {
                                 "append_results": False,
                                 "params": {
                                     "attempt_closest_match": False,
                                     "decimal_places": 1,
                                     "default_value": 0,
                                     "fallback_mode": "use_default",
                                     "fix_decimal_places": True,

                                     "is_unique": True,
                                     "lookalike_match": False,
                                     "maximum_value": 100.0,
                                     "minimum_value": 1.0,
                                     "skip_blank": True,
                                     "string_distance_threshold": 0.7
                                 },
                                 "rule_type": "Number",
                                 "field": "TotalEpisodes"
                             }]})

            date = "{0:%Y%m%d%H%M%S}".format(datetime.now())
            self.assertEqual(logs.output,
                             [
                                 'INFO:dativa.scrubber.test:modified: #26 from TotalEpisodes : Automatically fixed to 1 decimal places [totalepisodes-number-{0}.csv]'.format(
                                     date),
                                 'ERROR:dativa.scrubber.test:quarantined: #16 from TotalEpisodes_clean : Removed duplicates [totalepisodes_clean-number-{0}.csv]'.format(
                                     date),
                                 'WARNING:dativa.scrubber.test:replaced: #1 from TotalEpisodes : Replaced with default value [totalepisodes-number-{0}.csv]'.format(
                                     date)]
                             )

        shutil.rmtree(tempdir)

    def test_field_log_report_writer(self):
        csv = CSVHandler(base_path=self.base_path)
        logger = PersistentFieldLogger(logging.getLogger("dativa.scrubber.test"), {"id": 0})
        logger.debug(id=1)
        tempdir = tempfile.mkdtemp()

        df = csv.get_dataframe("number/test_int_is_unique_dirty.csv", force_dtype=pd.np.float)

        writer = LoggingReportWriter(logger=logger,
                                     quarantine_path=tempdir,
                                     error_severities=["quarantined"],
                                     warning_severities=["replaced"])
        scrubber = Scrubber(report_writer=writer)

        with self.assertLogs("dativa.scrubber.test") as logs:
            scrubber.run(df=df,
                         config={"rules": [
                             {
                                 "append_results": False,
                                 "params": {
                                     "attempt_closest_match": False,
                                     "decimal_places": 1,
                                     "default_value": 0,
                                     "fallback_mode": "use_default",
                                     "fix_decimal_places": True,

                                     "is_unique": True,
                                     "lookalike_match": False,
                                     "maximum_value": 100.0,
                                     "minimum_value": 1.0,
                                     "skip_blank": True,
                                     "string_distance_threshold": 0.7
                                 },
                                 "rule_type": "Number",
                                 "field": "TotalEpisodes"
                             }]})

            date = "{0:%Y%m%d%H%M%S}".format(datetime.now())
            self.assertEqual(logs.output,
                             ["INFO:dativa.scrubber.test:{{'id': 1, 'category': 'modified', 'field': 'TotalEpisodes', "
                              "'description': 'Automatically fixed to 1 decimal places', 'filename': "
                              "'totalepisodes-number-{0}.csv', 'number': 26}}".format(date),
                              "ERROR:dativa.scrubber.test:{{'id': 1, 'category': 'quarantined', 'field': "
                              "'TotalEpisodes_clean', 'description': 'Removed duplicates', 'filename': "
                              "'totalepisodes_clean-number-{0}.csv', 'number': 16}}".format(date),
                              "WARNING:dativa.scrubber.test:{{'id': 1, 'category': 'replaced', 'field': "
                              "'TotalEpisodes', 'description': 'Replaced with default value', 'filename': "
                              "'totalepisodes-number-{0}.csv', 'number': 1}}".format(date)])

        shutil.rmtree(tempdir)

    def test_lookalike(self):
        self._test_filter(dirty_file="generic/lookalike_dirty.csv",
                          clean_file="generic/lookalike_clean.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "",
                                          "fallback_mode": "remove_record",

                                          "is_unique": False,
                                          "lookalike_match": True,
                                          "maximum_length": 1000,
                                          "minimum_length": 0,
                                          "regex": ".+",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "String",
                                      "field": "-1"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'favourite_singer',
                                  'rule': 'String',
                                  'number_records': 2,
                                  'category': 'replaced',
                                  'description': 'Lookalike match',
                                  'df': [[None, 'Bob Dylan'],
                                         [None, 'Taylor Swift']],

                              },
                          ])

    def test_lookalike_2(self):
        self._test_filter(dirty_file="generic/lookalike_dirty2.csv",
                          clean_file="generic/lookalike_clean2.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "",
                                          "fallback_mode": "remove_record",

                                          "is_unique": False,
                                          "lookalike_match": True,
                                          "maximum_length": 1000,
                                          "minimum_length": 0,
                                          "regex": ".+",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "String",
                                      "field": "2"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'favourite_singer',
                                  'rule': 'String',
                                  'number_records': 3,
                                  'category': 'replaced',
                                  'description': 'Lookalike match',
                                  'df': [[None, 'Bob Dylan'],
                                         [None, 'Taylor Swift'],
                                         [None, 'David Bowie']],

                              },
                          ])

    def test_tsv(self):
        self._test_filter(dirty_file="generic/tsv_test.tsv",
                          clean_file="generic/tsv_clean2.tsv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "",
                                          "fallback_mode": "remove_record",

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_length": 1000,
                                          "minimum_length": 0,
                                          "regex": "^[0-9]*$",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "String",
                                      "field": "1"
                                  }
                              ]
                          },
                          csv_delimiter="\t",
                          report=[
                              {

                                  'field': 'name',
                                  'rule': 'String',
                                  'number_records': 6,
                                  'category': 'quarantined',
                                  'description': 'Data quarantined',
                                  'df': [['the good'],
                                         ['the bad'],
                                         ['the ugly'],
                                         ['the worse'],
                                         ['the best'],
                                         ['the coolest']],

                              },
                          ])

    def test_tsv_2(self):
        self._test_filter(dirty_file="generic/tsv_test.tsv",
                          clean_file="generic/tsv_clean.tsv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "default_value": "",
                                          "fallback_mode": "remove_record",

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_length": 1000,
                                          "minimum_length": 0,
                                          "regex": "^[0-9]*$",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "String",
                                      "field": "0"
                                  }
                              ]
                          },
                          csv_delimiter="\t",
                          report=[
                              {

                                  'field': 'count',
                                  'rule': 'String',
                                  'number_records': 2,
                                  'category': 'quarantined',
                                  'description': 'Data quarantined',
                                  'df': [[-1],
                                         [-2]],

                              },
                          ])

    def test_workflow_invalid_rule_name(self):
        self._test_filter(dirty_file="generic/tsv_test.tsv",
                          clean_file="generic/tsv_clean.tsv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "default_value": "",
                                          "fallback_mode": "remove_record",

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_length": 1000,
                                          "minimum_length": 0,
                                          "regex": "^[0-9]*$",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Banana",
                                      "field": "0"
                                  }
                              ]
                          },
                          expected_error=ScrubberValidationError)

    def test_missing_field(self):
        self._test_filter(dirty_file="generic/tsv_test.tsv",
                          clean_file="generic/tsv_clean.tsv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "default_value": "",
                                          "fallback_mode": "remove_record",

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "regex": "^[0-9]*$",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "String",
                                      "field": "0"
                                  }
                              ]
                          },
                          expected_error=ScrubberValidationError)

    def test_bad_multicolumn(self):
        self._test_filter(dirty_file="generic/tsv_test.tsv",
                          clean_file="generic/tsv_clean.tsv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "default_value": "",
                                          "fallback_mode": "remove_record",

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "regex": "^[0-9]*$",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Session",
                                  }
                              ]
                          },
                          expected_error=ScrubberValidationError)

    def test_missing_rules(self):
        self._test_filter(dirty_file="generic/tsv_test.tsv",
                          clean_file="generic/tsv_clean.tsv",
                          config={"no_rules": 1},
                          expected_error=ScrubberValidationError)


class WorkflowTestsWithRounding(_BaseTest):
    """
    these tests combine lookalike and rounding tests
    """

    def test_no_report(self):
        with self.assertRaises(ValueError):
            self._test_filter(dirty_file="generic/lookalike_short.csv",
                              clean_file="generic/lookalike_short_clean.csv",
                              config={
                                  "rules": [
                                      {
                                          "params": {
                                              "fix_decimal_places": True,
                                              "fallback_mode": "remove_record",
                                              "minimum_value": 20,
                                              "decimal_places": 0,
                                              "maximum_value": 100
                                          },
                                          "field": "Age",
                                          "rule_type": "Number"
                                      },
                                      {
                                          "params": {
                                              "maximum_length": 10,
                                              "minimum_length": 0,
                                              "fallback_mode": "remove_record"
                                          },
                                          "field": "Gender",
                                          "rule_type": "String"
                                      },
                                      {
                                          "params": {
                                              "maximum_length": 20,
                                              "skip_blank": False,
                                              "fallback_mode": "use_default",
                                              "lookalike_match": True,
                                              "is_unique": False,
                                              "minimum_length": 1,
                                              "attempt_closest_match": False,
                                              "default_value": "xxxxxxxxxxxxxxx"
                                          },
                                          "field": "Favourite artist",
                                          "rule_type": "String"
                                      }]})

    def test_lookalike_short(self):
        self._test_filter(dirty_file="generic/lookalike_short.csv",
                          clean_file="generic/lookalike_short_clean.csv",
                          config={
                              "rules": [
                                  {
                                      "params": {
                                          "fix_decimal_places": True,
                                          "fallback_mode": "remove_record",
                                          "minimum_value": 20,
                                          "decimal_places": 0,
                                          "maximum_value": 100
                                      },
                                      "field": "Age",
                                      "rule_type": "Number"
                                  },
                                  {
                                      "params": {
                                          "maximum_length": 10,
                                          "minimum_length": 0,
                                          "fallback_mode": "remove_record"
                                      },
                                      "field": "Gender",
                                      "rule_type": "String"
                                  },
                                  {
                                      "params": {
                                          "maximum_length": 20,
                                          "skip_blank": False,
                                          "fallback_mode": "use_default",
                                          "lookalike_match": True,
                                          "is_unique": False,
                                          "minimum_length": 1,
                                          "attempt_closest_match": False,
                                          "default_value": "xxxxxxxxxxxxxxx"
                                      },
                                      "field": "Favourite artist",
                                      "rule_type": "String"
                                  }]},
                          report=[
                              {

                                  'field': 'Age',
                                  'rule': 'Number',
                                  'number_records': 7,
                                  'category': 'modified',
                                  'description': 'Automatically fixed to 0 decimal places',
                                  'df': [[50, 'M', 'Bob Dylan', 50],
                                         [50, 'M', 'Bob Dylan', 50],
                                         [20, 'F', 'Taylor Swift', 20],
                                         [22, 'F', 'Rihanna', 22],
                                         [23, 'F', 'Rihanna', 23],
                                         [50, 'M', None, 50],
                                         [23, 'F', None, 23]],

                              },
                              {

                                  'field': 'Favourite artist',
                                  'rule': 'String',
                                  'number_records': 2,
                                  'category': 'replaced',
                                  'description': 'Lookalike match',
                                  'df': [[None, 'Bob Dylan'],
                                         [None, 'Rihanna']],

                              },
                          ]
                          )



# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
# Usage subject to license agreement
# hello@dativa.com for more information

from dativa.scrubber.tests import _BaseTest
from dativa.scrubber import ScrubberValidationError


class NumberTests(_BaseTest):

    def test_number(self):
        self._test_filter(dirty_file="number/test_int_range_dirty.csv",
                          clean_file="number/test_int_range_clean.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "decimal_places": 2,
                                          "default_value": 0,
                                          "fallback_mode": "use_default",
                                          "fix_decimal_places": False,

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_value": 1.0,
                                          "minimum_value": 0.0,
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Number",
                                      "field": "Number"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Number',
                                  'rule': 'Number',
                                  'number_records': 4,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [['1.02', 0],
                                         ['0', 0],
                                         ['-1', 0],
                                         ['fjidos', 0]],

                              },
                          ])

    def test_number_1(self):
        self._test_filter(dirty_file="number/test_int_range_dirty1.csv",
                          clean_file="number/test_int_range_clean1.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "decimal_places": 2,
                                          "default_value": 0,
                                          "fallback_mode": "use_default",
                                          "fix_decimal_places": False,

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_value": 1.0,
                                          "minimum_value": -1.0,
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Number",
                                      "field": "Number"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Number',
                                  'rule': 'Number',
                                  'number_records': 3,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [['1.02', 0],
                                         ['0', 0],
                                         ['fjidos', 0]],

                              },
                          ])

    def test_number_2(self):
        self._test_filter(dirty_file="number/test_int_range_dirty2.csv",
                          clean_file="number/test_int_range_clean2.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "decimal_places": 2,
                                          "default_value": 0,
                                          "fallback_mode": "use_default",
                                          "fix_decimal_places": False,

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_value": -1.0,
                                          "minimum_value": -5.0,
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Number",
                                      "field": "Number"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Number',
                                  'rule': 'Number',
                                  'number_records': 11,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [['-0.00', 0],
                                         ['-0.51', 0],
                                         ['-0.43', 0],
                                         ['0.00', 0],
                                         ['0.01', 0],
                                         ['0.02', 0],
                                         ['1.00', 0],
                                         ['1.02', 0],
                                         ['0', 0],
                                         ['fjidos', 0],
                                         ['jhjk', 0]],

                              },
                          ])

    def test_number_3(self):
        self._test_filter(dirty_file="number/test_int_range_dirty2.csv",
                          clean_file="number/test_int_range_clean2dp.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "decimal_places": 2,
                                          "default_value": 0,
                                          "fallback_mode": "use_default",
                                          "fix_decimal_places": False,

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_value": 10.0,
                                          "minimum_value": -10.0,
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Number",
                                      "field": "Number"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Number',
                                  'rule': 'Number',
                                  'number_records': 3,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [['0', 0],
                                         ['fjidos', 0],
                                         ['jhjk', 0]],

                              },
                          ]
                          )

    def test_number_4(self):
        self._test_filter(dirty_file="number/test_int_is_unique_dirty.csv",
                          clean_file="number/test_int_is_unique_clean.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "decimal_places": 1,
                                          "default_value": 0,
                                          "fallback_mode": "use_default",
                                          "fix_decimal_places": True,

                                          "is_unique": True,
                                          "lookalike_match": False,
                                          "maximum_value": 100.0,
                                          "minimum_value": 1.0,
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Number",
                                      "field": "TotalEpisodes"
                                  }
                              ]
                          },
                          report=[
                              {
                                  'field': 'TotalEpisodes',
                                  'rule': 'Number',
                                  'number_records': 26,
                                  'category': 'modified',
                                  'description': 'Automatically fixed to 1 decimal places',
                                  'df': [[1.3, 1.3],
                                         [19., 19.],
                                         [19., 19.],
                                         [19.001, 19.],
                                         [19.1, 19.1],
                                         [19.1, 19.1],
                                         [1.4, 1.4],
                                         [2.1, 2.1],
                                         [2.1, 2.1],
                                         [19.1, 19.1],
                                         [19.1, 19.1],
                                         [19.1, 19.1],
                                         [19.1, 19.1],
                                         [19.1, 19.1],
                                         [19.1, 19.1],
                                         [19.1, 19.1],
                                         [5., 5.],
                                         [5., 5.],
                                         [5.1, 5.1],
                                         [51.1, 51.1],
                                         [5.1, 5.1],
                                         [1.9, 1.9],
                                         [1.9, 1.9],
                                         [1.9, 1.9],
                                         [1., 1.],
                                         [1., 1.]]
                              },
                              {
                                  'field': 'TotalEpisodes_clean',
                                  'rule': 'Number',
                                  'number_records': 16,
                                  'category': 'quarantined',
                                  'description': 'Removed duplicates',
                                  'df': [[19., 19.],
                                         [19.001, 19.],
                                         [19.1, 19.1],
                                         [2.1, 2.1],
                                         [19.1, 19.1],
                                         [19.1, 19.1],
                                         [19.1, 19.1],
                                         [19.1, 19.1],
                                         [19.1, 19.1],
                                         [19.1, 19.1],
                                         [19.1, 19.1],
                                         [5., 5.],
                                         [5.1, 5.1],
                                         [1.9, 1.9],
                                         [1.9, 1.9],
                                         [1., 1.]]
                              },
                              {
                                  'field': 'TotalEpisodes',
                                  'rule': 'Number',
                                  'number_records': 1,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [[199.1, 0.]]
                              }]
                          )

    def test_number_5(self):
        self._test_filter(dirty_file="number/test_int_is_unique_dirty.csv",
                          clean_file="number/test_int_is_unique_clean2.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "decimal_places": 1,
                                          "default_value": 0,
                                          "fallback_mode": "use_default",
                                          "fix_decimal_places": True,

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_value": 100.0,
                                          "minimum_value": 1.0,
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Number",
                                      "field": "TotalEpisodes"
                                  }
                              ]
                          },
                          report=[{
                                 'field': 'TotalEpisodes',
                                 'rule': 'Number',
                                 'number_records': 26,
                                 'category': 'modified',
                                 'description': 'Automatically fixed to 1 decimal places',
                                 'df': [[1.3, 1.3],
                                        [19., 19.],
                                        [19., 19.],
                                        [19., 19.],
                                        [19.1, 19.1],
                                        [19.1, 19.1],
                                        [1.4, 1.4],
                                        [2.1, 2.1],
                                        [2.1, 2.1],
                                        [19.1, 19.1],
                                        [19.1, 19.1],
                                        [19.1, 19.1],
                                        [19.1, 19.1],
                                        [19.1, 19.1],
                                        [19.1, 19.1],
                                        [19.1, 19.1],
                                        [5., 5.],
                                        [5., 5.],
                                        [5.1, 5.1],
                                        [51.1, 51.1],
                                        [5.1, 5.1],
                                        [1.9, 1.9],
                                        [1.9, 1.9],
                                        [1.9, 1.9],
                                        [1., 1.],
                                        [1., 1.]]
                                 },
                                 {
                                      'field': 'TotalEpisodes',
                                      'rule': 'Number',
                                      'number_records': 1,
                                      'category': 'replaced',
                                      'description': 'Replaced with default value',
                                      'df': [[199.1, 0.]]
                                 }]
                          )

    def test_number_6(self):
        self._test_filter(dirty_file="number/test_int_range_dirty3.csv",
                          clean_file="number/test_int_range_clean3.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "decimal_places": 2,
                                          "fallback_mode": "remove_record",
                                          "fix_decimal_places": True,
                                          "maximum_value": 5.0,
                                          "minimum_value": -5.0,
                                          "skip_blank": True
                                      },
                                      "rule_type": "Number",
                                      "field": "Number"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Number',
                                  'rule': 'Number',
                                  'number_records': 13,
                                  'category': 'modified',
                                  'description': 'Automatically fixed to 2 decimal places',
                                  'df': [[0.0, '0.00'],
                                         [-0.51, '-0.51'],
                                         [-0.43, '-0.43'],
                                         [0.0, '0.00'],
                                         [0.01, '0.01'],
                                         [0.02, '0.02'],
                                         [1.0, '1.00'],
                                         [1.02, '1.02'],
                                         [0.0, '0.00'],
                                         [-1.0, '-1.00'],
                                         [-3.12, '-3.12'],
                                         [-1.32, '-1.32'],
                                         [-4.0, '-4.00']],

                              },
                          ])

    def test_number_7(self):
        self._test_filter(dirty_file="number/test_int_range_dirty_4.csv",
                          clean_file="number/test_int_range_clean_3.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "decimal_places": 2,
                                          "fallback_mode": "remove_record",
                                          "fix_decimal_places": False,
                                          "maximum_value": 5.0,
                                          "minimum_value": -5.0,
                                          "skip_blank": True
                                      },
                                      "rule_type": "Number",
                                      "field": "Number"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Number',
                                  'rule': 'Number',
                                  'number_records': 1,
                                  'category': 'quarantined',
                                  'description': 'Data quarantined',
                                  'df': [[0.001]],

                              },
                          ])

    def test_number_0dp(self):
        self._test_filter(dirty_file="number/test_int_range_dirty.csv",
                          clean_file="number/test_int_range_clean_0dp.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "decimal_places": 0,
                                          "default_value": 0,
                                          "fallback_mode": "use_default",
                                          "fix_decimal_places": True,

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_value": 1,
                                          "minimum_value": 0,
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Number",
                                      "field": "Number"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Number',
                                  'rule': 'Number',
                                  'number_records': 4,
                                  'category': 'modified',
                                  'description': 'Automatically fixed to 0 decimal places',
                                  'df': [['0.00', '0'],
                                         ['0.01', '0'],
                                         ['0.02', '0'],
                                         ['1.00', '1']],

                              },
                              {

                                  'field': 'Number',
                                  'rule': 'Number',
                                  'number_records': 3,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [['1.02', 0],
                                         ['-1', 0],
                                         ['fjidos', 0]],

                              },
                          ])

    def test_number_invalid(self):
        self._test_filter(dirty_file="number/test_int_range_dirty.csv",
                          clean_file="number/test_int_range_clean.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "decimal_places": 0,
                                          "default_value": 0,
                                          "fallback_mode": "invalid",
                                          "fix_decimal_places": False,

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_value": 1,
                                          "minimum_value": 0,
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Number",
                                      "field": "Number"
                                  }
                              ]
                          },
                          expected_error=ScrubberValidationError)

    def test_number_invalid_max_less_than_min(self):
        self._test_filter(dirty_file="number/test_int_range_dirty.csv",
                          clean_file="number/test_int_range_clean.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": False,
                                          "decimal_places": 0,
                                          "default_value": 0,
                                          "fallback_mode": "use_default",
                                          "fix_decimal_places": False,

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_value": 1,
                                          "minimum_value": 2,
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Number",
                                      "field": "Number"
                                  }
                              ]
                          },
                          expected_error=ScrubberValidationError)



# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
# Usage subject to license agreement
# hello@dativa.com for more information

from dativa.scrubber.tests import _BaseTest
from dativa.scrubber import ScrubberTooManyRecordsError, DefaultProfile
from dativa.tools.pandas import FpCSVEncodingError


class ProfileTests(_BaseTest):
    """Testing for user_subscriptions tests"""

    def test_profile(self):
        profile = DefaultProfile()

        self._test_filter(
            dirty_file="generic/Genres_Master.csv",
            clean_file="generic/Genres_Master.csv",
            config={"rules": [{"field": 0,
                               "rule_type": "String",
                               "params": {
                                   "regex": ".+",
                                   "fallback_mode": "use_default",
                                   "skip_blank": False,
                                   "minimum_length": 0,
                                   "maximum_length": 1024,
                                   "default_value": "N/A",
                                   "attempt_closest_match": False
                               }}]},
            profile=profile,
            report=[]
        )

        profile.maximum_records = 100

        self._test_filter(
            dirty_file="generic/Genres_Master.csv",
            clean_file="generic/Genres_Master.csv",
            config={"rules": [{"field": 0,
                               "rule_type": "String",
                               "params": {
                                   "regex": ".+",
                                   "fallback_mode": "use_default",
                                   "skip_blank": False,
                                   "minimum_length": 0,
                                   "maximum_length": 1024,
                                   "default_value": "N/A",
                                   "attempt_closest_match": False
                               }}]},
            profile=profile,
            expected_error=ScrubberTooManyRecordsError
        )

    def test_lookalike(self):
        profile = DefaultProfile()
        profile.maximum_records_lookalike = 1

        self._test_filter(dirty_file="generic/lookalike_dirty.csv",
                          clean_file="generic/lookalike_dirty.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "",
                                          "fallback_mode": "do_not_replace",

                                          "is_unique": False,
                                          "lookalike_match": True,
                                          "maximum_length": 1000,
                                          "minimum_length": 0,
                                          "regex": ".+",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "String",
                                      "field": "-1"
                                  }
                              ]
                          },
                          profile=profile,
                          report=[
                              {

                                  'field': 'favourite_singer',
                                  'rule': 'String',
                                  'number_records': 2,
                                  'category': 'ignored',
                                  'description': 'No changes made',
                                  'df': [[None, None],
                                         [None, None]],

                              },
                          ])

    def test_lookup_FpCSVEncodingError(self):
        self._test_filter(dirty_file="lookup/test_cities_dirty_windows1252.csv",
                          clean_file="lookup/test_cities_dirty_windows1252_deleted.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "BaNonea",
                                          "fallback_mode": "remove_record",
                                          "lookalike_match": False,
                                          "original_reference": "lookup/test_cities_reference1252.csv",

                                          "reference_field": "-1",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Lookup",
                                      "field": "1"
                                  }
                              ]
                          },
                          expected_error=FpCSVEncodingError,
                          report=[
                              {

                                  'field': 'favourite_singer',
                                  'rule': 'String',
                                  'number_records': 2,
                                  'category': 'ignored',
                                  'description': 'No changes made',
                                  'df': [[None, None],
                                         [None, None]],

                              },
                          ])

    def test_lookup_1252(self):
        self._test_filter(dirty_file="lookup/test_cities_dirty_windows1252.csv",
                          clean_file="lookup/test_cities_dirty_windows1252_deleted.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "BaNonea",
                                          "fallback_mode": "remove_record",
                                          "lookalike_match": False,
                                          "original_reference": "lookup/test_cities_reference1252.csv",

                                          "reference_field": "-1",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Lookup",
                                      "field": "1"
                                  }
                              ]
                          },
                          csv_encoding="Windows-1252",
                          report=[
                              {

                                  'field': 'city',
                                  'rule': 'Lookup',
                                  'number_records': 25,
                                  'category': 'replaced',
                                  'description': 'Best match',
                                  'df': [['Bsitrol', 'Bristol'],
                                         ['London', 'Croydon'],
                                         ['Lverpool', 'Liverpool'],
                                         ['méémlent', 'élément'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon']],

                              },
                              {

                                  'field': 'city',
                                  'rule': 'Lookup',
                                  'number_records': 3,
                                  'category': 'quarantined',
                                  'description': 'Data quarantined',
                                  'df': [['Bath'],
                                         ['Btha'],
                                         ['Lndon']],

                              },
                          ])



# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
# Usage subject to license agreement
# hello@dativa.com for more information

from dativa.scrubber.tests import _BaseTest
from dativa.scrubber import ScrubberError


class DuplicationTests(_BaseTest):

    def test_duplication(self):
        self._test_filter(dirty_file="unique/duplication_test.csv",
                          clean_file="unique/duplication_test_first_col.csv",
                          config={
                              "rules": [
                                  {
                                      "rule_type": "Uniqueness",
                                      "params": {
                                          "unique_fields": "Column1",
                                          "use_last_value": False
                                      },
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Column1',
                                  'rule': 'Uniqueness',
                                  'number_records': 2,
                                  'category': 'quarantined',
                                  'description': 'Removed duplicates',
                                  'df': [['A', 2],
                                         ['D', 4]],

                              },
                          ])

    def test_duplication_1(self):
        self._test_filter(dirty_file="unique/duplication_test.csv",
                          clean_file="unique/duplication_test_other_col.csv",
                          config={
                              "rules": [
                                  {
                                      "rule_type": "Uniqueness",
                                      "params": {
                                          "unique_fields": "Column2",
                                          "use_last_value": False
                                      },
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Column2',
                                  'rule': 'Uniqueness',
                                  'number_records': 2,
                                  'category': 'quarantined',
                                  'description': 'Removed duplicates',
                                  'df': [['C', 3],
                                         ['D', 4]],

                              },
                          ])

    def test_duplication_2(self):
        self._test_filter(dirty_file="unique/duplication_test.csv",
                          clean_file="unique/duplication_test_all_cols.csv",
                          config={
                              "rules": [
                                  {
                                      "rule_type": "Uniqueness",
                                      "params": {
                                          "unique_fields": "Column1,Column2",
                                          "use_last_value": False
                                      },
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Column1,Column2',
                                  'rule': 'Uniqueness',
                                  'number_records': 1,
                                  'category': 'quarantined',
                                  'description': 'Removed duplicates',
                                  'df': [['D', 4]],

                              },
                          ])

    def test_duplication_3(self):
        self._test_filter(dirty_file="unique/duplication_test.csv",
                          clean_file="unique/duplication_test_first_col_last.csv",
                          config={
                              "rules": [
                                  {
                                      "rule_type": "Uniqueness",
                                      "params": {
                                          "unique_fields": "Column1",
                                          "use_last_value": True
                                      },
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Column1',
                                  'rule': 'Uniqueness',
                                  'number_records': 2,
                                  'category': 'quarantined',
                                  'description': 'Removed duplicates',
                                  'df': [['A', 1],
                                         ['D', 4]],

                              },
                          ])

    def test_duplication_4(self):
        self._test_filter(dirty_file="unique/duplication_test.csv",
                          clean_file="unique/duplication_test_all_cols.csv",
                          config={
                              "rules": [
                                  {
                                      "rule_type": "Uniqueness",
                                      "params": {
                                          "unique_fields": "Column1,Column2",
                                          "use_last_value": False
                                      },
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Column1,Column2',
                                  'rule': 'Uniqueness',
                                  'number_records': 1,
                                  'category': 'quarantined',
                                  'description': 'Removed duplicates',
                                  'df': [['D', 4]],

                              },
                          ])

    def test_duplication_5(self):
        self._test_filter(dirty_file="unique/duplication_test.csv",
                          clean_file="unique/duplication_test_first_col.csv",
                          config={
                              "rules": [
                                  {
                                      "rule_type": "Uniqueness",
                                      "params": {
                                          "unique_fields": "Column1",
                                          "use_last_value": False
                                      },
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Column1',
                                  'rule': 'Uniqueness',
                                  'number_records': 2,
                                  'category': 'quarantined',
                                  'description': 'Removed duplicates',
                                  'df': [['A', 2],
                                         ['D', 4]],

                              },
                          ])

    def test_duplication_6(self):
        self._test_filter(dirty_file="unique/names_blank.csv",
                          clean_file="unique/names_blank_skip.csv",
                          config={
                              "rules": [
                                  {
                                      "rule_type": "Uniqueness",
                                      "params": {
                                          "unique_fields": "Name",
                                          "use_last_value": False
                                      },
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Name',
                                  'rule': 'Uniqueness',
                                  'number_records': 2,
                                  'category': 'quarantined',
                                  'description': 'Removed duplicates',
                                  'df': [['Sarvesh'],
                                         ['Sarvesh']],

                              },
                          ])

    def test_duplication_error(self):
        self._test_filter(dirty_file="unique/names_blank.csv",
                          clean_file="unique/names_blank_skip.csv",
                          config={
                              "rules": [
                                  {
                                      "rule_type": "Uniqueness",
                                      "params": {
                                          "unique_fields": "Donut",
                                          "use_last_value": False
                                      },
                                  }
                              ]
                          },
                          expected_error=ScrubberError)



import unittest
import logging
from os import path
import pandas as pd
from collections import OrderedDict
import json
from dativa.tools.pandas import CSVHandler
from dativa.scrubber import DefaultProfile
from dativa.analyzer import AutoConfig

logger = logging.getLogger("dativa.analyzer.tests")


def _compare_objects(new_object, original_object, ignore_list=[]):
    if original_object is None:
        return new_object, None
    if new_object is None:
        return original_object, None
    for key in original_object:
        if key not in ignore_list:
            if key in new_object:
                if type(original_object[key]) == dict or type(original_object[key]) == OrderedDict:
                    a, b = _compare_objects(
                        new_object[key], original_object[key], ignore_list)
                    if a != b:
                        return a, b
                elif type(original_object[key]) == list:
                    for index in range(0, len(original_object[key])):
                        a, b = _compare_objects(
                            new_object[key][index], original_object[key][index], ignore_list)
                        if a != b:
                            return a, b
                elif type(original_object[key]) == tuple:
                    # skip tuples
                    pass
                elif original_object[key] is None:
                    return key, None
                else:
                    if json.dumps(original_object[key], sort_keys=True) != json.dumps(new_object[key], sort_keys=True):
                        return json.dumps("{0}: {1}".format(key, original_object[key]), sort_keys=True), json.dumps(
                            "{0}: {1}".format(key, new_object[key]), sort_keys=True)
            else:
                return key, None

    return True, True


def compare_objects(new_object, original_object, ignore_list=[]):
    a, b = _compare_objects(new_object, original_object, ignore_list)
    if a is True and b is True:
        return _compare_objects(original_object, new_object, ignore_list)
    else:
        return a, b


class AutoConfigTests(unittest.TestCase):

    def setUp(self):
        self.base_path = "{0}/test-data/analyzer/".format(
            path.dirname(path.abspath(__file__)))
        self.csv = CSVHandler(base_path=self.base_path)
        super(AutoConfigTests, self).setUp()

    def _test(self, file, config, description, iteration=10, profile=None):
        """
        Iterative test that will run up to five times to check for success.
        This is because the tests are not-deterministic due to the pandas
        sampling approach
        """
        df = self.csv.get_dataframe(file, force_dtype=pd.np.str)

        ac = AutoConfig()

        auto_config, df_dict = ac.create_config(df, profile)

        a, b = compare_objects(auto_config, config)

        if a != b and iteration > 0:
            logger.debug(
                "executed iteration #{iteration}, re-running".format(iteration=11 - iteration))
            self._test(file, config, description, iteration - 1)
            return
        if a != b:
            import pdb;
            pdb.set_trace()
        self.assertEquals(a, b)

        if description != ac.describe_dataframe(auto_config, df_dict):
            self.assertEquals(
                description, ac.describe_dataframe(auto_config, df_dict))

    def test_autoplan_blank_test(self):

        config = {"rules": [{
            "field": "word",
            "rule_type": "String",
            "params": {
                "minimum_length": 1,
                "maximum_length": 10,
                "fallback_mode": "use_default",
                "default_value": "",
                "skip_blank": True,
                "is_unique": True}},
            {"field": "number",
             "rule_type": "Number",
             "params": {
                 "minimum_value": 0,
                 "maximum_value": 0,
                 "decimal_places": 0,
                 "fix_decimal_places": True,
                 "fallback_mode": "remove_record"
             }}]}

        description = "column,type,minimum,maximum,format,has_blanks\n" + \
                      "word,String,1,10,,True\n" + \
                      "number,Number,0,0,int,False\n\n"

        self._test("blank_test.csv", config,
                   description, profile=DefaultProfile())

    def test_autoplan_blank_dirty(self):

        config = {"rules":
                      [{"field": "word",
                        "rule_type": "String",
                        "params": {
                            "minimum_length": 1,
                            "maximum_length": 10,
                            "fallback_mode": "use_default",
                            "skip_blank": True,
                            "default_value": "",
                            "skip_blank": True}},
                       {"field": "number",
                        "rule_type": "Number",
                        "params": {
                            "minimum_value": 0,
                            "maximum_value": 0,
                            "decimal_places": 0,
                            "fix_decimal_places": True,
                            "fallback_mode": "remove_record",
                            "skip_blank": True
                        }}]}

        description = "column,type,minimum,maximum,format,has_blanks\n" + \
                      "word,String,1,10,,True\n" + \
                      "number,Number,0,0,int,True\n\n"

        self._test("blank_dirty.csv", config, description)

    def test_autoplan_cities(self):
        config = {"rules": [{
            "field": "id",
            "rule_type": "Number",
            "params": {
                "minimum_value": 1,
                "maximum_value": 10,
                "decimal_places": 0,
                "fix_decimal_places": True,
                "fallback_mode": "remove_record",
                "is_unique": True}},
            {"field": "city",
             "rule_type": "String",
             "params": {
                 "minimum_length": 0,
                 "maximum_length": 10,
                 "fallback_mode": "use_default",
                 "default_value": "Bath",
                 "is_unique": True}}]}

        description = "column,type,minimum,maximum,format,has_blanks\n" + \
                      "id,Number,1,10,int,False\n" + \
                      "city,String,0,10,,False\n\n"

        self._test("test_cities_dirty.csv", config, description)

    def test_positive_dirty(self):

        config = {"rules": [{
            "field": "Column1",
            "rule_type": "Number",
            "params": {
                "minimum_value": 0,
                "maximum_value": 10,
                "decimal_places": 0,
                "fix_decimal_places": True,
                "fallback_mode": "remove_record"}},
            {"field": "Column2",
             "rule_type": "Number",
             "params": {
                 "minimum_value": 1,
                 "maximum_value": 10,
                 "decimal_places": 0,
                 "fix_decimal_places": True,
                 "fallback_mode": "remove_record"
             }}]}

        description = "column,type,minimum,maximum,format,has_blanks\n" + \
                      "Column1,Number,0,10,int,False\n" + \
                      "Column2,Number,1,10,int,False\n\n"

        self._test("positive_dirty.csv", config, description)

    def test_autoplan_fridge(self):

        config = {"rules": [{
            "field": "device_id",
            "rule_type": "Number",
            "params": {
                "minimum_value": 1000,
                "maximum_value": 20000000,
                "decimal_places": 0,
                "fix_decimal_places": True,
                "fallback_mode": "remove_record"}},
            {"field": "session_start",
             "rule_type": "Date",
             "params": {
                 "date_format": "%d/%m/%Y %H:%M",
                 'range_check': 'fixed',
                 "range_minimum": "01/01/2016 00:00",
                 "range_maximum": "31/12/2016 23:59",
                 "fallback_mode": "remove_record"}},
            {"field": "session_stop",
             "rule_type": "Date",
             "params": {
                 "date_format": "%d/%m/%Y %H:%M",
                 'range_check': 'fixed',
                 "range_minimum": "01/01/2016 00:00",
                 "range_maximum": "31/12/2016 23:59",
                 "fallback_mode": "remove_record"}},
            {"field": "average_temperature",
             "rule_type": "Number",
             "params": {
                 "decimal_places": 2,
                 "fix_decimal_places": True,
                 "minimum_value": 1,
                 "maximum_value": 10,
                 "decimal_places": 2,
                 "fix_decimal_places": True,
                 "fallback_mode": "remove_record"}},
            {"rule_type": "Uniqueness",
             "params": {
                 "unique_fields": "device_id,session_start",
                 "use_last_value": True}},
            {"rule_type": "Session",
             "params": {
                 "key_field": "device_id",
                 "start_field": "session_start",
                 "end_field": "session_stop",
                 "gaps_option": "extend_end",
                 "overlaps_option": "truncate_end",
                 "date_format": "%d/%m/%Y %H:%M"
             }}]}

        description = "column,type,minimum,maximum,format,has_blanks\n" + \
                      "device_id,Number,1000,20000000,int,False\n" + \
                      "session_start,Date,01/01/2016 00:00,31/12/2016 23:59,%d/%m/%Y %H:%M,False\n" + \
                      "session_stop,Date,01/01/2016 00:00,31/12/2016 23:59,%d/%m/%Y %H:%M,False\n" + \
                      "average_temperature,Number,1,10,0.00,False\n" + \
                      "\n" + \
                      "Unique on device_id,session_start\n" + \
                      "Session rule on session_start, session_stop with unique id device_id\n"

        self._test("fridge-test-small.csv", config, description)

    def test_session_checking(self):

        config = {
            "rules": [
                {
                    "field": "Channel",
                    "rule_type": "String",
                    "params": {
                        "minimum_length": 1,
                        "maximum_length": 10,
                        'fallback_mode': 'remove_record',
                        'regex': 'HBO'
                    }
                },
                {
                    "field": "Show",
                    "rule_type": "String",
                    "params": {
                        "minimum_length": 0,
                        "maximum_length": 20,
                        "fallback_mode": "use_default",
                        "default_value": "Game of Thrones"
                    }
                },
                {
                    "field": "Start",
                    "rule_type": "Date",
                    "params": {
                        "date_format": "%Y-%m-%d %H:%M:%S",
                        'range_check': 'fixed',
                        "range_minimum": "2016-12-01 00:00:00",
                        "range_maximum": "2016-12-01 23:59:59",
                        "fallback_mode": "remove_record",
                        "is_unique": True
                    }
                },
                {
                    "field": "Stop",
                    "rule_type": "Date",
                    "params": {
                        "date_format": "%Y-%m-%d %H:%M:%S",
                        'range_check': 'fixed',
                        "range_minimum": "2016-12-01 00:00:00",
                        "range_maximum": "2016-12-01 23:59:59",
                        "fallback_mode": "remove_record"
                    }
                },
                {
                    "rule_type": "Session",
                    "params": {
                        "key_field": "Show",
                        "start_field": "Start",
                        "end_field": "Stop",
                        "gaps_option": "ignore",
                        "overlaps_option": "truncate_end",
                        "date_format": "%Y-%m-%d %H:%M:%S"
                    }
                }
            ]
        }

        description = "column,type,minimum,maximum,format,has_blanks\n" + \
                      "Channel,String,1,10,,False\n" + \
                      "Show,String,0,20,,False\n" + \
                      "Start,Date,2016-12-01 00:00:00,2016-12-01 23:59:59,%Y-%m-%d %H:%M:%S,False\n" + \
                      "Stop,Date,2016-12-01 00:00:00,2016-12-01 23:59:59,%Y-%m-%d %H:%M:%S,False\n" + \
                      "\n" + \
                      "Session rule on Start, Stop with unique id Show\n"

        self._test("session_checking.csv", config, description)

    def test_epg(self):

        config = {
            'rules': [
                {'field': 'station_name',
                 'rule_type': 'Lookup',
                 'params': {
                     'original_reference': 'station_name.ref.csv',
                     'reference_field': 0,
                     'attempt_closest_match': True,
                     'fallback_mode': 'remove_record'}},
                {'field': 'airdate',
                 'rule_type': 'Date',
                 'params': {
                     'date_format': '%Y-%m-%d %H:%M:%S',
                     'range_check': 'fixed',
                     'range_minimum': '2017-01-01 00:00:00',
                     'range_maximum': '2017-02-28 23:59:59',
                     'fallback_mode': 'remove_record'}},
                {'field': 'title',
                 'rule_type': 'String',
                 'params': {
                     'minimum_length': 1,
                     'maximum_length': 100,
                     'fallback_mode': 'remove_record'}},
                {'field': 'epi_num',
                 'rule_type': 'String',
                 'params': {
                     'minimum_length': 1,
                     'maximum_length': 10,
                     'fallback_mode': 'use_default',
                     'skip_blank': True,
                     'default_value': ''}},
                {'field': 'epi_title',
                 'rule_type': 'String',
                 'params': {
                     'minimum_length': 1,
                     'maximum_length': 200,
                     'fallback_mode': 'use_default',
                     'skip_blank': True,
                     'default_value': ''}},
                {'field': 'genre',
                 'rule_type': 'String',
                 'params': {
                     'minimum_length': 1,
                     'maximum_length': 100,
                     'fallback_mode': 'remove_record'}},
                {'field': 'year',
                 'rule_type': 'Number',
                 'params': {
                     'decimal_places': 0,
                     'fix_decimal_places': True,
                     'minimum_value': 1000,
                     'maximum_value': 3000,
                     'fallback_mode': 'remove_record',
                     'skip_blank': True}},
                {'field': 'language',
                 'rule_type': 'String',
                 'params': {
                     'minimum_length': 0,
                     'maximum_length': 10,
                     'fallback_mode': 'remove_record',
                     'regex': 'English|Spanish'}},
                {'field': 'country_name',
                 'rule_type': 'Lookup',
                 'params': {
                     'original_reference': 'country_name.ref.csv',
                     'reference_field': 0,
                     'attempt_closest_match': True,
                     'fallback_mode': 'remove_record',
                     'skip_blank': True}},
                {'rule_type': 'Uniqueness',
                 'params': {
                     'unique_fields': 'station_name,airdate',
                     'use_last_value': True}}]}

        description = "column,type,minimum,maximum,format,has_blanks\n" + \
                      "station_name,String,A & E Network,truTV HD,Lookup,False\n" + \
                      "airdate,Date,2017-01-01 00:00:00,2017-02-28 23:59:59,%Y-%m-%d %H:%M:%S,False\n" + \
                      "title,String,1,100,,False\n" + \
                      "epi_num,String,1,10,,True\n" + \
                      "epi_title,String,1,200,,True\n" + \
                      "genre,String,1,100,,False\n" + \
                      "year,Number,1000,3000,int,True\n" + \
                      "language,String,0,10,,False\n" + \
                      "country_name,String,AUS/TUR/USA,ZAF/USA,Lookup,True\n" + \
                      "\n" + \
                      "Unique on station_name,airdate\n"

        self._test("epg-jan-2017.csv", config, description)



# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
# Usage subject to license agreement
# hello@dativa.com for more information

from dativa.scrubber.tests import _BaseTest
from dativa.scrubber import ScrubberValidationError


class StringTests(_BaseTest):

    def test_to_text(self):
        bucket = "test-bucket"

        business_rules = [
            {'rule_type': 'Date',
             'field': 'timestamp',
             'params': {'fallback_mode': 'remove_record',
                        'skip_blank': False,
                        'date_format': '%Y-%m-%d %H:%M:%S'}},
            {'rule_type': 'String',
             'field': 'discount',
             'params': {'fallback_mode': 'remove_record',
                        'minimum_length': 0,
                        'maximum_length': 1}},
            {'rule_type': 'String',
             'field': 'custState',
             'params': {'fallback_mode': 'remove_record',
                        'minimum_length': 0,
                        'maximum_length': 2,
                        'skip_blank': True,
                        'regex': '(?!NY|CA)'}},
            {'rule_type': 'String',
             'field': 'expressConsent',
             'params': {'fallback_mode': 'remove_record',
                        'minimum_length': 4,
                        'skip_blank': True,
                        'maximum_length': 4,
                        'regex': '(T|t)rue'}},
            {'rule_type': 'Number',
             'field': 'latitude',
             'params': {'fallback_mode': 'remove_record',
                        'fix_decimal_places': True,
                        'decimal_places': 7,
                        'minimum_value': -90,
                        'maximum_value': 90}},
            {'rule_type': 'Number',
             'field': 'longitude',
             'params': {'fallback_mode': 'remove_record',
                        'fix_decimal_places': True,
                        'decimal_places': 7,
                        'minimum_value': -180,
                        'maximum_value': 180}},
            {'rule_type': 'String',
             'field': 'custCountry',
             'params': {'maximum_length': 3,
                        'minimum_length': 2,
                        'regex': 'US.?',
                        'fallback_mode': 'remove_record',
                        'skip_blank': True}},

            {'rule_type': 'String',
             'field': 'vin',
             'params': {'maximum_length': 17,
                        'minimum_length': 17,
                        'regex': '.{12}\d{5}',
                        'fallback_mode': 'remove_record',
                        }},
        ]
        source_params = {'minimum_length': 2,
                         'maximum_length': 5,
                         'regex': 'IDS2|GM|TOY|TOYO',
                         'fallback_mode': 'remove_record'}

        source_params_blanks = {'minimum_length': 2,
                                'maximum_length': 5,
                                'skip_blank': True,
                                'regex': 'IDS2|GM|TOY|ZCW',
                                'fallback_mode': 'remove_record'}

        date_param_blank = {'fallback_mode': 'remove_record',
                            'skip_blank': True,
                            'date_format': '%Y-%m-%d %H:%M:%S'}

        cleaning_rules = [
            {'rule_type': 'String',
             'field': 'engineOff',
             'params': {'minimum_length': 0,
                        'maximum_length': 10,
                        'regex': 'True|False',
                        'skip_blank': True,
                        'fallback_mode': 'remove_record'}},
            {'rule_type': 'Number',
             'field': 'fuelCapacity',
             'params': {'minimum_value': 0,
                        'maximum_value': 200,
                        'fallback_mode': 'remove_record',
                        'decimal_places': 1,
                        'fix_decimal_places': True}},
            {'rule_type': 'String',
             'field': 'fuelUom',
             'params': {'minimum_length': 0,
                        'maximum_length': 1024,
                        'regex': 'gal',
                        'fallback_mode': 'remove_record'}},
            {'rule_type': 'String',
             'field': 'gpsCountry',
             'params': {'minimum_length': 0,
                        'maximum_length': 25,
                        'fallback_mode': 'remove_record',
                        'skip_blank': True}},
            {'rule_type': 'Date',
             'field': 'lastCommDate',
             'params': date_param_blank},
            {'rule_type': 'Date',
             'field': 'lastUpdate',
             'params': date_param_blank},
            {'rule_type': 'String',
             'field': 'licenseCountry',
             'params': {'fallback_mode': 'remove_record',
                        'minimum_length': 2,
                        'maximum_length': 2}},
            {'rule_type': 'String',
             'field': 'make',
             'params': {'fallback_mode': 'remove_record',
                        'maximum_length': 4,
                        'minimum_length': 2}},
            {'rule_type': 'String',
             'field': 'model',
             'params': {'fallback_mode': 'remove_record',
                        'maximum_length': 4,
                        'minimum_length': 4}},
            {'rule_type': 'String',
             'field': 'mva',
             'params': {'fallback_mode': 'remove_record',
                        'maximum_length': 10,
                        'minimum_length': 8}},
            {'rule_type': 'String',
             'field': 'odometerUom',
             'params': {'maximum_length': 2,
                        'minimum_length': 2,
                        'fallback_mode': 'remove_record'}},
            {'rule_type': 'Date',
             'field': 'parkedTime',
             'params': date_param_blank},
            {'rule_type': 'String',
             'field': 'source',
             'params': source_params},
            {'rule_type': 'String',
             'field': 'tcu',
             'params': source_params},
            {'rule_type': 'Number',
             'field': 'year',
             'params': {'maximum_value': 2030,
                        'minimum_value': 2000,
                        'fallback_mode': 'remove_record',
                        'decimal_places': 0,
                        'fix_decimal_places': False}},
            {'rule_type': 'String',
             'field': 'source_fuelMetrics',
             'params': source_params_blanks},
            {'rule_type': 'String',
             'field': 'source_geoPoint',
             'params': source_params_blanks},
            {'rule_type': 'String',
             'field': 'source_odometerMetrics',
             'params': source_params_blanks},
            {'rule_type': 'String',
             'field': 'tcu_tireMetrics',
             'params': source_params_blanks},
            {'rule_type': 'String',
             'field': 'tcu_fuelMetrics',
             'params': source_params_blanks},
            {'rule_type': 'String',
             'field': 'tcu_tcuData',
             'params': source_params_blanks},
            {'rule_type': 'String',
             'field': 'tcu_geoPoint',
             'params': source_params_blanks},
            {'rule_type': 'String',
             'field': 'tcu_odometerMetrics',
             'params': source_params_blanks},
            {'rule_type': 'String',
             'field': 'tcu_rentalData',
             'params': source_params_blanks},
            {'rule_type': 'Date',
             'field': 'timestamp_tireMetrics',
             'params': date_param_blank},
            {'rule_type': 'Date',
             'field': 'timestamp_fuelMetrics',
             'params': date_param_blank},
            {'rule_type': 'Date',
             'field': 'timestamp_geoPoint',
             'params': date_param_blank},
            {'rule_type': 'Date',
             'field': 'timestamp_odometerMetrics',
             'params': date_param_blank},
            {'rule_type': 'Date',
             'field': 'timestamp_rentalData',
             'params': date_param_blank},
            {'rule_type': 'Date',
             'field': 'checkInDue',
             'params': date_param_blank},
            {'rule_type': 'Number',
             'field': 'fuel',
             'params': {'maximum_value': 200,
                        'minimum_value': 0,
                        'decimal_places': 2,
                        'fix_decimal_places': True,
                        'skip_blank': True,
                        'fallback_mode': 'remove_record'}},
            {'rule_type': 'Number',
             'field': 'fuelInEights',
             'params': {'maximum_value': 8,
                        'minimum_value': 0,
                        'skip_blank': True,
                        'decimal_places': 0,
                        'fix_decimal_places': True,
                        'fallback_mode': 'remove_record'}},
            {'rule_type': 'Number',
             'field': 'fuelInTenths',
             'params': {'maximum_value': 200,
                        'minimum_value': 0,
                        'decimal_places': 1,
                        'skip_blank': True,
                        'fix_decimal_places': True,
                        'fallback_mode': 'remove_record'}},
            {'rule_type': 'String',
             'field': 'uom_fuelMetrics',
             'params': {'maximum_length': 3,
                        'minimum_length': 1,
                        'skip_blank': True,
                        'fallback_mode': 'remove_record',
                        'regex': 'gal|l'}},
            {'rule_type': 'Number',
             'field': 'odometer',
             'params': {'maximum_value': 200000,
                        'minimum_value': 0,
                        'decimal_places': 1,
                        'fix_decimal_places': True,
                        'skip_blank': True,
                        'fallback_mode': 'remove_record'}},
            {'rule_type': 'String',
             'field': 'uom_odometerMetrics',
             'params': {'maximum_length': 2,
                        'minimum_length': 0,
                        'fallback_mode': 'remove_record',
                        'skip_blank': True,
                        'regex': 'mi|km'}},
            {'rule_type': 'String',
             'field': 'brand',
             'params': {'maximum_length': 1,
                        'minimum_length': 0,
                        'regex': 'A|B|C|P',
                        'skip_blank': True,
                        'fallback_mode': 'remove_record'}},
            {'rule_type': 'String',
             'field': 'ciLocIdentifier',
             'params': {'maximum_length': 1,
                        'minimum_length': 0,
                        'regex': 'C|A|L',
                        'skip_blank': True,
                        'fallback_mode': 'remove_record'}},
            {'rule_type': 'String',
             'field': 'coLocIdentifier',
             'params': {'maximum_length': 1,
                        'minimum_length': 0,
                        'regex': 'C|A|L',
                        'skip_blank': True,
                        'fallback_mode': 'remove_record'}},
            {'rule_type': 'String',
             'field': 'rentalStatus',
             'params': {'maximum_length': 6,
                        'minimum_length': 6,
                        'fallback_mode': 'remove_record',
                        'skip_blank': True}}
        ]

        config = dict()

        config['rules'] = business_rules + cleaning_rules

        scrubber = self.ScrubberClass()

        self.assertEquals(scrubber.to_text(config),
                          """field,type,date_format,decimal_places,fallback_mode,fix_decimal_places,maximum_length,maximum_value,minimum_length,minimum_value,regex,skip_blank
timestamp,Date,%Y-%m-%d %H:%M:%S,,remove_record,,,,,,,
discount,String,,,remove_record,,1.0,,0.0,,,
custState,String,,,remove_record,,2.0,,0.0,,(?!NY|CA),True
expressConsent,String,,,remove_record,,4.0,,4.0,,(T|t)rue,True
latitude,Number,,7.0,remove_record,True,,90.0,,-90.0,,
longitude,Number,,7.0,remove_record,True,,180.0,,-180.0,,
custCountry,String,,,remove_record,,3.0,,2.0,,US.?,True
vin,String,,,remove_record,,17.0,,17.0,,.{12}\d{5},
engineOff,String,,,remove_record,,10.0,,0.0,,True|False,True
fuelCapacity,Number,,1.0,remove_record,True,,200.0,,0.0,,
fuelUom,String,,,remove_record,,1024.0,,0.0,,gal,
gpsCountry,String,,,remove_record,,25.0,,0.0,,,True
lastCommDate,Date,%Y-%m-%d %H:%M:%S,,remove_record,,,,,,,True
lastUpdate,Date,%Y-%m-%d %H:%M:%S,,remove_record,,,,,,,True
licenseCountry,String,,,remove_record,,2.0,,2.0,,,
make,String,,,remove_record,,4.0,,2.0,,,
model,String,,,remove_record,,4.0,,4.0,,,
mva,String,,,remove_record,,10.0,,8.0,,,
odometerUom,String,,,remove_record,,2.0,,2.0,,,
parkedTime,Date,%Y-%m-%d %H:%M:%S,,remove_record,,,,,,,True
source,String,,,remove_record,,5.0,,2.0,,IDS2|GM|TOY|TOYO,
tcu,String,,,remove_record,,5.0,,2.0,,IDS2|GM|TOY|TOYO,
year,Number,,0.0,remove_record,False,,2030.0,,2000.0,,
source_fuelMetrics,String,,,remove_record,,5.0,,2.0,,IDS2|GM|TOY|ZCW,True
source_geoPoint,String,,,remove_record,,5.0,,2.0,,IDS2|GM|TOY|ZCW,True
source_odometerMetrics,String,,,remove_record,,5.0,,2.0,,IDS2|GM|TOY|ZCW,True
tcu_tireMetrics,String,,,remove_record,,5.0,,2.0,,IDS2|GM|TOY|ZCW,True
tcu_fuelMetrics,String,,,remove_record,,5.0,,2.0,,IDS2|GM|TOY|ZCW,True
tcu_tcuData,String,,,remove_record,,5.0,,2.0,,IDS2|GM|TOY|ZCW,True
tcu_geoPoint,String,,,remove_record,,5.0,,2.0,,IDS2|GM|TOY|ZCW,True
tcu_odometerMetrics,String,,,remove_record,,5.0,,2.0,,IDS2|GM|TOY|ZCW,True
tcu_rentalData,String,,,remove_record,,5.0,,2.0,,IDS2|GM|TOY|ZCW,True
timestamp_tireMetrics,Date,%Y-%m-%d %H:%M:%S,,remove_record,,,,,,,True
timestamp_fuelMetrics,Date,%Y-%m-%d %H:%M:%S,,remove_record,,,,,,,True
timestamp_geoPoint,Date,%Y-%m-%d %H:%M:%S,,remove_record,,,,,,,True
timestamp_odometerMetrics,Date,%Y-%m-%d %H:%M:%S,,remove_record,,,,,,,True
timestamp_rentalData,Date,%Y-%m-%d %H:%M:%S,,remove_record,,,,,,,True
checkInDue,Date,%Y-%m-%d %H:%M:%S,,remove_record,,,,,,,True
fuel,Number,,2.0,remove_record,True,,200.0,,0.0,,True
fuelInEights,Number,,0.0,remove_record,True,,8.0,,0.0,,True
fuelInTenths,Number,,1.0,remove_record,True,,200.0,,0.0,,True
uom_fuelMetrics,String,,,remove_record,,3.0,,1.0,,gal|l,True
odometer,Number,,1.0,remove_record,True,,200000.0,,0.0,,True
uom_odometerMetrics,String,,,remove_record,,2.0,,0.0,,mi|km,True
brand,String,,,remove_record,,1.0,,0.0,,A|B|C|P,True
ciLocIdentifier,String,,,remove_record,,1.0,,0.0,,C|A|L,True
coLocIdentifier,String,,,remove_record,,1.0,,0.0,,C|A|L,True
rentalStatus,String,,,remove_record,,6.0,,6.0,,,True
""")



# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
#  This code is licensed under MIT license (see license.txt for details)

#!python
#cython: language_level=3, boundscheck=False, wraparound=False, optimize.use_switch=True
cimport numpy as np
import numpy as np
import Levenshtein

"""
Contains optimized string distance functions for the best matches and lookalike matches
Contains .pxd file to compile with cython
"""

# def get_distance(word_1, word_2):
cdef inline double get_distance(str word_1, str word_2):
    cdef double d
    cdef int length_1
    """
    Returns an estimate of the similarity of strings,
    where 1.0 means they are identical an 0 means they
    are totally disimillar.

    For single words, uses the jaro_winkler distance,
    Otherwise use metric based on the Leveshtein distance
    """

    length_1 = len(word_1)

    if length_1 > 15:
        d = Levenshtein.distance(word_1, word_2) / length_1
        if d > 1:
            return 0
        else:
            return 1 - d
    else:
        return Levenshtein.jaro_winkler(word_1, word_2)

# def which_pass(cycle, string_1, string_2, string_1_length, string_2_length):
cdef inline int which_pass(int cycle, str string_1, str string_2, int string_1_length, int string_2_length):
    """
    Both the best matches and the lookalike match implement a multi-pass process
    as follows:
    1. Match only strings of comparable length of more than 1 character in length or
       where the first and second letters match
    2. Match strings of comparable length where the first letter matches
    3. Match strings where the first letter matches
    4. Match strings where the length is comparable
    5. Match all others that are not more that twice the length of each other

    Parameters:
    cycle: the cycle to check eligbility against
    string_1: the first string
    string_2: the second string
    string_1_length: the optional length of the first string
    string_2_length: the optional length fo the second string
    """

    if cycle == 1:
        if string_2_length == 1 and string_1_length == 1:
            return True
        else:
            return abs(string_2_length - string_1_length) < 3 and string_2[0] == string_1[0] and string_2[1] == \
                   string_1[1]
    elif cycle == 2:
        return abs(string_2_length - string_1_length) < 3 and string_2[0] == string_1[0]

    elif cycle == 3:
        return string_2[0] == string_1[0]

    elif cycle == 4:
        return abs(string_2_length - string_1_length) < 3

    elif cycle == 5:
        return string_1_length < string_2_length * 2 and string_2_length < string_1_length * 2
    else:
        return False

# def which_pass_no_length(cycle, string_1, string_2):
cdef inline int which_pass_no_length(int cycle, str string_1, str string_2):
    """
    Provides a wrapper or which_pass when no string length is cached
    """
    return which_pass(cycle, string_1, string_2, len(string_1), len(string_2))

# def get_closest_match(string_list, possible_values, threshold=0.7):
cpdef np.ndarray get_closest_match(np.ndarray[object, ndim=1] string_list, np.ndarray[object, ndim=1] possible_values,
                                   double threshold=0.7):
    cdef np.ndarray[object, ndim=1] results  # noqa
    cdef np.ndarray[np.float_t, ndim=1] distances  # noqa
    cdef np.ndarray[np.float_t, ndim=1] ref_distances  # noqa
    cdef np.ndarray[np.int_t, ndim=1] ref_lengths  # noqa
    cdef Py_ssize_t string_ix
    cdef Py_ssize_t possible_ix
    cdef Py_ssize_t string_length
    cdef Py_ssize_t target_ix
    cdef int cycle

    """
    Takes a list containing strings to be matched against a set of possible
    values and returns a matrstring_ix matching the two to the closest values,
    provided the string similarity is less than the specified threshold

    Parameters
    ----------

    string_list : the list containing the strings to be compared
    possible_values : an array of possible values the string can take
    threshold: the string similarity threshold

    """

    results = np.zeros(len(string_list), dtype=object)
    distances = np.zeros((len(string_list)), dtype=np.float)
    ref_distances = np.zeros((len(possible_values)), dtype=np.float)
    ref_lengths = np.zeros((len(possible_values)), dtype=np.int)

    for string_ix in range(0, len(possible_values)):
        ref_lengths[string_ix] = len(possible_values[string_ix])

    for string_ix in range(0, len(string_list)):
        string_length = len(string_list[string_ix])

        # cycle through the reference list in four passes
        for cycle in [1, 2, 3, 4, 5]:
            # continue only while we don't have match that is above the
            # threshold
            if distances[string_ix] < threshold:
                # cycle through every row and calculate the distance
                ref_distances.fill(0)
                for possible_ix in range(0, len(possible_values)):
                    if which_pass(cycle,
                                  possible_values[possible_ix],
                                  string_list[string_ix],
                                  ref_lengths[possible_ix],
                                  string_length):
                        ref_distances[possible_ix] = get_distance(possible_values[possible_ix], string_list[string_ix])

                # get the closest match
                target_ix = np.argsort(ref_distances)[len(possible_values) - 1]
                results[string_ix] = possible_values[target_ix]
                distances[string_ix] = ref_distances[target_ix]

    return np.vstack((results, distances))

# def run_lookalike(matrix, rows_to_match=None, top_n=5, threshold=0.7):
cpdef np.ndarray run_lookalike(np.ndarray matrix,
                               np.ndarray rows_to_match,
                               int top_n=5,
                               double threshold=0.7):
    cdef np.ndarray reference
    cdef np.ndarray choices
    cdef np.ndarray uniques
    cdef np.ndarray unique_counts
    cdef np.ndarray sort
    cdef Py_ssize_t result_column
    cdef Py_ssize_t distance_column
    cdef Py_ssize_t match_column
    cdef Py_ssize_t first_column
    cdef Py_ssize_t last_column
    cdef Py_ssize_t number_rows
    cdef int cycle
    cdef int row_pass
    cdef Py_ssize_t number_uniques
    """
    Runs a lookalike match on the passed matrix and returns as an array from
    the final column in the matrix. The lookalike column works by scanning
    through all comparable columns looking for the top n most similar columns.
    It then counts the frequency of each result and chooses the most popular.

    Parameters
    ----------
    matrix:  A numpy matrix containing:
        - the strings to be used for the lookalike match
        - the values of the target field
        - a blank column of int64, which will contain the distances
        - a blank column of strings, the will contain the result
    rows_to_match: an array specifying which rows need to be matched
    top_n: the number of matches to be compared
    threshold: the string similarity threshold for comparisons
    """
    # useful variable for tracking the columns in the matrix
    result_column = matrix.shape[0] - 1
    distance_column = matrix.shape[0] - 2
    match_column = matrix.shape[0] - 3
    first_column = 0

    # the last column contains the distance calculation
    last_column = matrix.shape[0] - 3
    number_rows = matrix.shape[1]

    # iterate through every row in the matrix
    for ix in range(0, number_rows):

        # if do not need to match this
        if rows_to_match is None or rows_to_match[ix]:
            # grab the reference row
            reference = np.transpose(matrix)[ix]

            # cycle through every row and calculate the distance
            for cycle in [1, 2, 3, 5]:
                if matrix[distance_column, matrix[distance_column] > threshold * (last_column - first_column)].shape[
                    0] - 1 < top_n:
                    for row_ix in range(0, number_rows):
                        if (rows_to_match is not None and rows_to_match[row_ix]) or row_ix == ix:
                            matrix[distance_column, row_ix] = 100
                        else:
                            row_pass = 0
                            for col_ix in range(first_column, last_column):
                                row_pass = row_pass + \
                                           which_pass_no_length(cycle,
                                                                reference[col_ix],
                                                                matrix[col_ix][row_ix])

                            if cycle == round(row_pass / (last_column - first_column)):
                                total_distance = 0
                                for col_ix in range(first_column, last_column):
                                    d = get_distance(
                                        reference[col_ix], matrix[col_ix][row_ix])
                                    total_distance = total_distance + \
                                                     (1 - d) * (1 - d)

                                matrix[distance_column,
                                       row_ix] = total_distance

            # choose amongst the most popular of the top_n neighbours
            choices = matrix[match_column][np.argsort(matrix[distance_column])[:top_n]]
            uniques, unique_counts = np.unique(choices, return_counts=True)
            number_uniques = len(uniques)
            if number_uniques > 1:
                sort = unique_counts.argsort()
                if unique_counts[sort[number_uniques - 1]] > unique_counts[sort[number_uniques - 2]]:
                    matrix[result_column, ix] = uniques[sort][number_uniques - 1]
                else:
                    for i in range(0, len(choices)):
                        if choices[i] == uniques[sort[number_uniques - 1]] or choices[i] == uniques[
                            sort[number_uniques - 2]]:
                            matrix[result_column, ix] = choices[i]
                            break
            elif number_uniques == 1:
                matrix[result_column, ix] = uniques[0]

    return matrix[result_column]



from .base import DefaultProfile, ScrubberError, ScrubberTooManyRecordsError, ScrubberValidationError

from .scrubber import Scrubber
from .report import DefaultReportWriter, LoggingReportWriter, PersistentFieldLogger

__all__ = ['DefaultProfile',
           'Scrubber',
           'ScrubberError', 'ScrubberTooManyRecordsError','ScrubberValidationError',
           'DefaultReportWriter', 'LoggingReportWriter', 'PersistentFieldLogger']



# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
# Usage subject to license agreement
# hello@dativa.com for more information

import pandas as pd
from dativa.tools.pandas import CSVHandler
from dativa.scrubber.tests import _BaseTest
from dativa.scrubber import Scrubber


class SkipBlankTests(_BaseTest):

    def test_skip_blank(self):
        self._test_filter(dirty_file="generic/names_blank.csv",
                          clean_file="generic/names_blank_skip.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "N/A",
                                          "fallback_mode": "use_default",

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_length": 1000,
                                          "minimum_length": 0,
                                          "regex": ".+",
                                          "skip_blank": True,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "String",
                                      "field": "0"
                                  }
                              ]
                          },
                          report=[])

    def test_skip_blank_1(self):
        self._test_filter(dirty_file="generic/names_blank.csv",
                          clean_file="generic/names_blank_na.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "N/A",
                                          "fallback_mode": "use_default",

                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_length": 1000,
                                          "minimum_length": 0,
                                          "regex": ".+",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "String",
                                      "field": "0"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'Name',
                                  'rule': 'String',
                                  'number_records': 1,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [[None, 'N/A']],

                              },
                          ])

    def test_skip_blank_2(self):
        self._test_filter(dirty_file="generic/test_cities_dirty_skip_blank.csv",
                          clean_file="generic/test_cities_dirty_skip_blank_na.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "Banana",
                                          "fallback_mode": "do_not_replace",
                                          "lookalike_match": False,
                                          "original_reference": "generic/test_cities_reference.csv",

                                          "reference_field": "-1",
                                          "skip_blank": False,
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "Lookup",
                                      "field": "1"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'city',
                                  'rule': 'Lookup',
                                  'number_records': 23,
                                  'category': 'replaced',
                                  'description': 'Best match',
                                  'df': [['Bsitrol', 'Bristol'],
                                         ['Lverpool', 'Liverpool'],
                                         ['méémlent', 'élément'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon'],
                                         ['London', 'Croydon']],

                              },
                              {

                                  'field': 'city',
                                  'rule': 'Lookup',
                                  'number_records': 8,
                                  'category': 'ignored',
                                  'description': 'No changes made',
                                  'df': [['Bath', 'Bath'],
                                         ['Btha', 'Btha'],
                                         ['Lndon', 'Lndon'],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None],
                                         [None, None]],

                              },
                          ])

    def test_skip_blank_3(self):
        self._test_filter(dirty_file="generic/email_test.csv",
                          clean_file="generic/email_test_skip_blank.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "N/A",
                                          "fallback_mode": "use_default",
                                          "is_unique": False,
                                          "lookalike_match": False,
                                          "maximum_length": 1000,
                                          "minimum_length": 0,
                                          "regex": ".+",
                                          "string_distance_threshold": 0.7
                                      },
                                      "rule_type": "String",
                                      "field": "0"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'e-mail',
                                  'rule': 'String',
                                  'number_records': 1,
                                  'category': 'replaced',
                                  'description': 'Replaced with default value',
                                  'df': [[None, 'N/A']],

                              },
                          ])

    def test_skip_blank_native(self):
        csv = CSVHandler(base_path=self.base_path)

        df = csv.get_dataframe("number/test_int_is_unique_dirty.csv", force_dtype=pd.np.float)

        fp = Scrubber()

        fp.run(df=df,
               config={"rules": [
                   {
                       "append_results": False,
                       "params": {
                           "attempt_closest_match": False,
                           "decimal_places": 1,
                           "default_value": 0,
                           "fallback_mode": "use_default",
                           "fix_decimal_places": True,

                           "is_unique": True,
                           "lookalike_match": False,
                           "maximum_value": 100.0,
                           "minimum_value": 1.0,
                           "skip_blank": True,
                           "string_distance_threshold": 0.7
                       },
                       "rule_type": "Number",
                       "field": "TotalEpisodes"
                   }]})

        self._compare_df_to_file(csv, df, "number/test_int_is_unique_clean.csv")

    def test_skip_blank_columns(self):
        self._test_filter(dirty_file="generic/email_test.csv",
                          clean_file="generic/email_missing_field.csv",
                          config={
                              "rules": [
                                  {
                                      "append_results": False,
                                      "params": {
                                          "attempt_closest_match": True,
                                          "default_value": "N/A",
                                          "fallback_mode": "use_default",
                                          "maximum_length": 1000,
                                          "minimum_length": 0,
                                      },
                                      "rule_type": "String",
                                      "field": "not_a_field"
                                  }
                              ]
                          },
                          report=[
                              {

                                  'field': 'not_a_field',
                                  'rule': 'String',
                                  'number_records': 7,
                                  'category': 'quarantined',
                                  'description': 'File did not contain required column',
                                  'df': [['n@gmail.com', 0, None],
                                         ['p@gmail.com', 1, None],
                                         ['h@gmail.com', 0, None],
                                         ['s@gmail.com', 0, None],
                                         ['l@gmail.com', 1, None],
                                         ['v@gmail.com', 0, None],
                                         [None, 0, None]],

                              },
                          ])



# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
#  This code is licensed under MIT license (see license.txt for details)

import datetime
import pandas as pd
from dativa.tools.pandas import CSVHandler


class ReportEntry:
    """
    An individual report entry.

    Parameters
    ----------
    rule: the name of the rule that produced this entry
    field: the field being processed when the entry was produced
    df: the DataFrame containing the invalid records
    category: the type of entry, one of:
        HistoryCategory.IGNORED - the records were not altered
        HistoryCategory.REPLACED - the records were replaced with other values
        HistoryCategory.REMOVED - the records have been removed
        HistoryCategory.MODIFIED - the records have been modified
        HistoryCategory.INSERTED - new records have been inserted
    description: a textual description of the changes made (if any)
    """

    date = ""
    field = ""
    df = None
    category = ""
    description = ""
    number_records = ""

    def __init__(self, rule, field, df, category, description):
        self.date = datetime.datetime.now()
        self.field = field
        self.df = pd.DataFrame(df)
        self.number_records = df.shape[0]
        self.category = category
        self.description = description
        self.rule = rule

    def __str__(self):
        return "{0}, Field {1}({2}): #{3} {4}/|{5}".format(self.date,
                                                           self.field,
                                                           self.rule,
                                                           self.number_records,
                                                           self.category,
                                                           self.description)

    def get_log_dict(self):
        """
        Returns a dict containing the log entries
        """
        return {"date": self.date,
                "field": self.field,
                "rule": self.rule,
                "number_records": self.number_records,
                "category": self.category,
                "description": self.description}

    def get_records(self):
        """
        returns the DataFrame containing the values
        """
        return self.df


class DefaultReportWriter:
    """
    A default implementation of the ReportWriter that stores ReportEntries
    and returns them.
    """

    report = []

    def reset_report(self):
        self.report = []

    def log_history(self,
                    rule,
                    field,
                    df,
                    category,
                    description
                    ):
        """
        Logs a report entry
        """

        assert isinstance(df, pd.DataFrame), "df passed to log_history is not a DataFrame"

        self.report.append(ReportEntry(rule=rule,
                                       field=field,
                                       df=df,
                                       category=category,
                                       description=description
                                       ))

        return

    def get_report(self):
        """
        Returns an array of all ReportEntries created
        """
        return self.report


class PersistentFieldLogger:

    def __init__(self, logger, persistent_fields):
        self.logger = logger
        self.persistent_fields = persistent_fields

    def _get_args(self, **kwargs):
        args = self.persistent_fields.copy()
        for arg in kwargs:
            if arg in self.persistent_fields:
                self.persistent_fields[arg] = kwargs[arg]
            args[arg] = kwargs[arg]

        return args

    def debug(self, **kwargs):
        self.logger.debug(self._get_args(**kwargs))

    def info(self, **kwargs):
        self.logger.info(self._get_args(**kwargs))

    def warning(self, **kwargs):
        self.logger.warning(self._get_args(**kwargs))

    def error(self, **kwargs):
        self.logger.error(self._get_args(**kwargs))


class LoggingReportWriter(DefaultReportWriter):
    """
    A class to log all reports to the python log file
    """

    def __init__(self, logger, quarantine_path, error_severities=None, warning_severities=None):
        self.logger = logger
        self.csv = CSVHandler(base_path=quarantine_path)
        self.error_severities = [] if error_severities is None else error_severities
        self.warning_severities = [] if warning_severities is None else warning_severities

    def log_history(self,
                    rule,
                    field,
                    df,
                    category,
                    description
                    ):
        """
        Logs out the report details and saves the dataframe to the specified path
        """

        if df is not None and df.shape[0] > 0:
            filename = "{field}-{rule}-{date:%Y%m%d%H%M%S}.csv".format(
                field=field,
                rule=rule,
                date=datetime.datetime.now()).lower()
            self.csv.save_df(df, filename)

            if type(self.logger) is PersistentFieldLogger:
                if category in self.error_severities:
                    self.logger.error(category=category,
                                      field=field,
                                      description=description,
                                      filename=filename,
                                      number=df.shape[0])
                elif category in self.warning_severities:
                    self.logger.warning(category=category,
                                        field=field,
                                        description=description,
                                        filename=filename,
                                        number=df.shape[0])
                else:
                    self.logger.info(category=category,
                                     field=field,
                                     description=description,
                                     filename=filename,
                                     number=df.shape[0])
            else:
                message = "{category}: #{number} from {field} : {description} [{filename}]".format(
                    category=category,
                    field=field,
                    description=description,
                    filename=filename,
                    number=df.shape[0]
                )

                if category in self.error_severities:
                    self.logger.error(message)
                elif category in self.warning_severities:
                    self.logger.warning(message)
                else:
                    self.logger.info(message)



from .base_test import _BaseTest
from .test_append import AppendTests
from .test_anonymization import AnonymizationTests
from .test_best_match import BestMatchTests
from .test_date import DateTests
from .test_lookup import LookupTests
from .test_number import NumberTests
from .test_profile import ProfileTests
from .test_session import SessionTests
from .test_skip_blank import SkipBlankTests
from .test_string import StringTests
from .test_unique import DuplicationTests
from .test_workflow import WorkflowTests, WorkflowTestsWithRounding

__all__ = ["_BaseTest",
           "AppendTests",
           "AnonymizationTests",
           "BestMatchTests",
           "DateTests",
           "LookupTests",
           "NumberTests",
           "ProfileTests",
           "SessionTests",
           "SkipBlankTests",
           "StringTests",
           "DuplicationTests",
           "WorkflowTests",
           "WorkflowTestsWithRounding"]



# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
#  This code is licensed under MIT license (see license.txt for details)

import logging
import datetime
import hashlib
from Crypto.PublicKey import RSA
from Crypto.Cipher import PKCS1_OAEP
from base64 import b64encode, b64decode
from dativa.tools.pandas import get_unique_column_name, datetime_to_string, string_to_datetime, format_string_is_valid
from .base import BaseAddIn, ScrubberValidationError, HistoryType

logger = logging.getLogger("dativa.scrubber.anonymization")

def _validate_key(key):
    try:
        RSA.importKey(key)
        return True
    except ValueError:
        return False


class EncryptionAddIn(BaseAddIn):
    """
    An addin that can be added to any BaseValidator class to provide encryption of the field
    """

    fields = {"encrypt": {"default": False, "validator": lambda x: x in [True, False]},
              "public_key": {"default": "none", "validator": _validate_key},
              "random_string": {"default": "random", "validator": lambda x: len(x) < 21},
              "random_string_date_field": {"default": "none", "validator": lambda x: True},
              "random_string_date_format": {"default": "none", "validator": format_string_is_valid},
              }

    def cross_validation(self):
        """
        Provides cross-validation for the fields in the encryption add-in
        """
        if self.validator.encrypt is True:
            if self.validator.public_key == "none":
                raise ScrubberValidationError("The param public_key must be set if encrypt is true")
            if self.validator.random_string_date_field != "none":
                if self.validator.random_string_date_format == "none":
                    raise ScrubberValidationError(
                        "random_string_date_format must be set is random_string_date_field is set")
                if not format_string_is_valid(self.validator.random_string):
                    raise ScrubberValidationError(
                        "{0} is not a valid value for random_string. This must be date parseable if "
                        "random_string_date_field is set".format(self.validator.random_string))

    @staticmethod
    def _pad(value, length):
        if len(value) > length:
            raise ScrubberValidationError(
                "random {0} is too long, should be no more than {1} bytes".format(value, length))
        return value + b'\00' * (length - len(value))

    def _get_encrypted(self, value, cipher, random):
        bytes_to_encrypt = str.encode(value, encoding="UTF-8")
        if random != b'random':
            # Update the random function directly within the cipher
            # as we want it to change with every record
            cipher._randfunc = lambda x: self._pad(random, x)

        encrypted_bytes = cipher.encrypt(bytes_to_encrypt)
        return str(b64encode(encrypted_bytes), encoding="UTF-8")

    def _encrypt(self, df, column):
        """
        Encrypts the passed column using RSA encryption with AOEP
        """
        public_key = RSA.importKey(self.validator.public_key)
        cipher = PKCS1_OAEP.new(public_key)

        # set up additional columns we will use
        cols = dict()
        cols["random_number"] = get_unique_column_name(df, "random", "_random")

        # get the random_string for each row
        if self.validator.random_string_date_field == "none":
            df[cols["random_number"]] = str.encode(self.validator.random_string, 'UTF-8')
        else:
            if self.validator.random_string_date_field not in df.columns:
                raise ScrubberValidationError(
                    "The random_string_date_field %s was not present in the passed DataFrame".format(
                        self.validator.random_string_date_field))
            else:
                dates = string_to_datetime(
                    df[self.validator.random_string_date_field], format=self.validator.random_string_date_format,
                    errors='coerce')
                df[cols["random_number"]] = datetime_to_string(dates, self.validator.random_string)
                df[cols["random_number"]] = df[cols["random_number"]].apply(lambda x: str.encode(x, 'UTF-8'))

        df[column] = df.apply(lambda row: self._get_encrypted(row[column], cipher, row[cols["random_number"]]), axis=1)

        if self.validator.random_string_date_field == "none":
            self.report_writer.log_history(
                self.validator.rule_type,
                column,
                df[column].to_frame(),
                HistoryType.MODIFIED,
                "Encrypted values")
        else:
            self.report_writer.log_history(
                self.validator.rule_type,
                column,
                df[[column, self.validator.random_string_date_field]],
                HistoryType.MODIFIED,
                "Encrypted values to date field")

        df.drop(cols["random_number"], axis=1, inplace=True)

    def run(self, df, column, original_column=None):
        """
        Runs the encryption add-in on the passed column
        """
        if self.validator.encrypt:
            self._encrypt(df, column)


class DecryptionAddIn(BaseAddIn):
    """
    An addin that can be added to any BaseValidator class to provide decryption of the field
    """

    fields = {"decrypt": {"default": False, "validator": lambda x: x in [True, False]},
              "private_key": {"default": "none", "validator": _validate_key}
              }

    def cross_validation(self):
        """
        Provides cross-validation for the decryption add-in
        """
        if self.validator.decrypt is True:
            if self.validator.private_key == "none":
                raise ScrubberValidationError("The param private_key must be set if decrypt is true")

    @staticmethod
    def _get_decrypted(value, cipher):
        bytes_to_decrypt = b64decode(str.encode(value, encoding="UTF-8"))
        decrypted_bytes = cipher.decrypt(bytes_to_decrypt)
        return str(decrypted_bytes, encoding="UTF-8")

    def _decrypt(self, df, column):
        """
        Decrypts the passed column using RSA encryption with AOEP
        """
        private_key = RSA.importKey(self.validator.private_key)
        cipher = PKCS1_OAEP.new(private_key)
        df[column] = df[column].apply(lambda message: self._get_decrypted(message, cipher))

        self.report_writer.log_history(
            self.validator.rule_type,
            column,
            df[column].to_frame(),
            HistoryType.MODIFIED,
            "Decrypted all values")

    def run(self, df, column, original_column=None):
        """
        Runs the decryption add-ing
        """
        if self.validator.decrypt:
            self._decrypt(df, column)


class HashingAddIn(BaseAddIn):
    """
    An addin that can be added to any BaseValidator class to provide hashing of the field
    """

    fields = {"hash": {"default": False, "validator": lambda x: x in [True, False]},
              "hash_length": {"default": 16, "validator": lambda x: 0 < x < 129},
              "salt": {"default": "none", "validator": lambda x: len(x) < 17},
              "salt_date_field": {"default": "none", "validator": lambda x: True},
              "salt_date_format": {"default": "none", "validator": format_string_is_valid},
              }

    def cross_validation(self):
        """
        Provides cross-validation on the fields passed to the hashing add-in
        """
        if self.validator.hash is True:
            if self.validator.salt_date_field != "none":
                if self.validator.salt_date_format != "none":
                    if format_string_is_valid(self.validator.salt):
                        salt = datetime_to_string(datetime.datetime.now(), self.validator.salt)
                        if len(salt) > 16:
                            raise ScrubberValidationError(
                                "Salt {0} is too long once the data is applied, maximum length 16 bytes".format(
                                    self.validator.salt))
                    else:
                        raise ScrubberValidationError(
                            "{0} is not a valid value for salt. This must be date parseable if salt_date_field is set".format(
                                self.validator.salt))
                else:
                    raise ScrubberValidationError("salt_date_format must be set if salt_date_field is set")

    def _get_hash(self, value, salt):
        """
        Gets the hash for the passed value with the specific salt
        """
        return hashlib.sha512(value + salt).hexdigest()[0:self.validator.hash_length]

    def _hash(self, df, column):
        """
        Hashes the passed column with the salt in the fields passed to the add-in
        """

        cols = dict()
        cols["salt"] = get_unique_column_name(df, "salt")
        cols["bytes"] = get_unique_column_name(df, "bytes")

        if self.validator.salt_date_field == "none":
            df[cols["salt"]] = str.encode(self.validator.salt, 'UTF-8')
        else:
            if self.validator.salt_date_field not in df.columns:
                raise ScrubberValidationError("The salt_date_field %s was not present in the passed DataFrame".format(
                    self.validator.salt_date_field))
            else:
                dates = string_to_datetime(df[self.validator.salt_date_field], format=self.validator.salt_date_format,
                                           errors='coerce')
                df[cols["salt"]] = datetime_to_string(dates, self.validator.salt).apply(
                    lambda x: str.encode(x, 'UTF-8'))

        df[cols["bytes"]] = df[column].apply(lambda x: str.encode(x, 'UTF-8'))
        df[column] = df.apply(lambda row: self._get_hash(row[cols["bytes"]], row[cols["salt"]]), axis=1)

        df.drop(cols["salt"], axis=1, inplace=True)
        df.drop(cols["bytes"], axis=1, inplace=True)

        if self.validator.salt_date_field == "none":
            self.report_writer.log_history(
                self.validator.rule_type,
                column,
                df[column].to_frame(),
                HistoryType.MODIFIED,
                "Hashed all values")
        else:
            self.report_writer.log_history(
                self.validator.rule_type,
                column,
                df[[self.validator.salt_date_field, column]],
                HistoryType.MODIFIED,
                "Hashed all values to salt date")

    def run(self, df, column, original_column=None):
        """
        Runs the hashing add-ing
        """
        if self.validator.hash:
            self._hash(df, column)



# (c) 2012-2018 Dativa, all rights reserved
# -----------------------------------------
# Usage subject to license agreement
# hello@dativa.com for more information

import unittest
import logging
# import json
from os import path, remove
import pandas as pd
from dativa.tools.pandas import CSVHandler
from dativa.scrubber import Scrubber

logger = logging.getLogger("dativa.scrubber.tests")


class _BaseTest(unittest.TestCase):
    ScrubberClass = Scrubber

    def setUp(self):
        self.base_path = "{0}/test-data/".format(
            path.dirname(path.abspath(__file__)))
        super(_BaseTest, self).setUp()

    def _compare_df_to_file(self, csv, df, file):
        """

        :param csv: an instance of a CSVHandler class
        :param df: the dataframe to compare
        :param file: the file to compare
        :return: raises an exception if they not the same
        """

        # save and reload the df to overcome differences in file handling
        # between platforms
        temp_file = "temp.csv"
        csv.save_df(df, temp_file)
        test_df = csv.load_df(temp_file)
        remove(path.join(csv.base_path, temp_file))

        clean_df = csv.load_df(file)
        pd.testing.assert_frame_equal(clean_df, test_df)

    def _test_filter_actual(self,
                            dirty_file,
                            clean_file,
                            config,
                            expected_report,
                            csv_encoding,
                            csv_delimiter,
                            csv_header,
                            csv_skiprows,
                            csv_quotechar,
                            report_writer,
                            profile,
                            expected_error,
                            expected_df_dict,
                            force_dtype):

        csv = CSVHandler(csv_encoding=csv_encoding,
                         csv_delimiter=csv_delimiter,
                         csv_header=csv_header,
                         csv_skiprows=csv_skiprows,
                         csv_quotechar=csv_quotechar,
                         base_path=self.base_path,
                         force_dtype=str)

        scrubber = self.ScrubberClass(report_writer, profile)

        dataframe_dict = {}
        for file in scrubber.get_files_from_config(config):
            dataframe_dict[file] = csv.get_dataframe(file, force_dtype=pd.np.str)

        df = csv.get_dataframe(dirty_file, force_dtype=force_dtype)
        report = scrubber.run(df, config, dataframe_dict)

        # -------------------------------------------------------------------------
        # UNCOMMENT THIS CODE TO ASSIST IN GETTING THE REPORT OBJECTS FOR NEW TESTS
        # -------------------------------------------------------------------------
        # print(",")
        # print("report = [")
        # import numpy as np
        # np.set_printoptions(threshold=np.nan)
        # for r in report:
        #     print("{")
        #     print("""
        #     'field': '{0}',
        #            'rule': '{1}',
        #            'number_records': {2},
        #            'category': '{3}',
        #            'description': '{4}',
        #            'df': {5},
        #            """.format(r.field, r.rule, r.number_records, r.category, r.description,
        #                       repr(r.df.values).replace("array(", "").replace(", dtype=object)", "").replace(
        #                           "dtype=object),", "").replace("Timestamp", "_to_time")).replace("nan", "None"))
        #     print("},")
        # print("]")
        # -------------------------------------------------------------------------

        if expected_report is not None:

            self.assertEqual(len(report), len(expected_report))
            for i in range(0, len(report)):
                self.assertEqual(report[i].field, expected_report[i]["field"])
                self.assertEqual(report[i].rule, expected_report[i]["rule"])
                self.assertEqual(report[i].number_records, expected_report[i]["number_records"])
                self.assertEqual(report[i].category, expected_report[i]["category"])
                self.assertEqual(report[i].description, expected_report[i]["description"])
                pd.np.array_equal(report[i].df.values, expected_report[i]["df"], )

        self._compare_df_to_file(csv, df, clean_file)

        if expected_df_dict is not None:
            for file in expected_df_dict:
                try:
                    expected_df = csv.get_dataframe(expected_df_dict[file])
                    actual_df = dataframe_dict[file]
                    pd.testing.assert_frame_equal(expected_df, actual_df)
                except TypeError:
                    expected_df = csv.get_dataframe(file)
                    actual_df = dataframe_dict[file]
                    pd.testing.assert_frame_equal(expected_df, actual_df)

    def _test_filter(self,
                     dirty_file,
                     clean_file,
                     config,
                     report=None,
                     csv_encoding="UTF-8",
                     csv_delimiter=",",
                     csv_header=0,
                     csv_skiprows=0,
                     csv_quotechar='"',
                     report_writer=None,
                     profile=None,
                     expected_error=None,
                     expected_df_dict=None,
                     force_dtype=None):

        logger.debug("testing {0}".format(dirty_file))

        if expected_error is None:
            val = self._test_filter_actual(dirty_file,
                                           clean_file,
                                           config,
                                           report,
                                           csv_encoding,
                                           csv_delimiter,
                                           csv_header,
                                           csv_skiprows,
                                           csv_quotechar,
                                           report_writer,
                                           profile,
                                           expected_error,
                                           force_dtype=force_dtype,
                                           expected_df_dict=expected_df_dict)

            if report is None:
                raise ValueError("Report not set for test")

            return val

        with self.assertRaises(expected_error):
            return self._test_filter_actual(dirty_file,
                                            clean_file,
                                            config,
                                            report,
                                            csv_encoding,
                                            csv_delimiter,
                                            csv_header,
                                            csv_skiprows,
                                            csv_quotechar,
                                            report_writer,
                                            profile,
                                            expected_error,
                                            force_dtype=force_dtype,
                                            expected_df_dict=expected_df_dict)
